from fastapi import FastAPI
import socketio

def socketio_mount(
    app: FastAPI,
    async_mode: str = "asgi",
    mount_path: str = "/socket.io/",
    socketio_path: str = "socket.io",
    logger: bool = False,
    engineio_logger: bool = False,
    cors_allowed_origins="*",
    **kwargs
) -> socketio.AsyncServer:
    """Mounts an async SocketIO app over an FastAPI app."""

    sio = socketio.AsyncServer(async_mode=async_mode,
                      cors_allowed_origins=cors_allowed_origins,
                      logger=logger,
                      engineio_logger=engineio_logger, **kwargs)

    sio_app = socketio.ASGIApp(sio, socketio_path=socketio_path)

    # mount
    app.add_route(mount_path, route=sio_app, methods=["GET", "POST"]) # type: ignore
    app.add_websocket_route(mount_path, sio_app) # type: ignore

    return sio


from orca.system.thread_manager import IThreadManager
from orca.workflow_models.status_enums import MethodStatus
from orca.workflow_models.labware_thread import IMethodObserver, Method
from orca.workflow_models.workflow_templates import ThreadTemplate


class SpawnThreadAction(IMethodObserver):
    def __init__(self, 
                 thread_manager: IThreadManager, 
                 thread_template: ThreadTemplate) -> None:
        self._thread_manager: IThreadManager = thread_manager
        self._template: ThreadTemplate = thread_template
        self._has_executed: bool = False

    def execute(self) -> None:
        thread = self._thread_manager.create_thread_instance(self._template)
        self._has_executed = True

    def method_notify(self, event: str, method: Method) -> None:
        if self._has_executed:
            return
        if event == MethodStatus.IN_PROGRESS.name.upper():
            self._template.set_wrapped_method(method)
            self.execute()


from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Dict, List
from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.system.move_handler import MoveHandler
from orca.system.reservation_manager import IReservationManager
from orca.system.system_map import SystemMap
from orca.workflow_models.status_enums import LabwareThreadStatus
from orca.workflow_models.labware_thread import LabwareThread
from orca.workflow_models.workflow_templates import ThreadTemplate


orca_logger = logging.getLogger("orca")

class ThreadFactory:
    def __init__(self, labware_registry: ILabwareRegistry, move_handler: MoveHandler, reservation_manager: IReservationManager, system_map: SystemMap) -> None:
        self._labware_registry: ILabwareRegistry = labware_registry
        self._system_map: SystemMap = system_map
        self._move_handler = move_handler
        self._reservation_manager = reservation_manager

    def create_instance(self, template: ThreadTemplate) -> LabwareThread:

        # Instantiate labware
        labware_instance = template.labware_template.create_instance()
        self._labware_registry.add_labware(labware_instance)

        # create the thread
        thread = LabwareThread(labware_instance,
                                template.start_location,
                                template.end_location,
                                self._move_handler,
                                self._reservation_manager,
                                self._system_map,
                                template.observers)
        return thread


class IThreadRegistry(ABC):
    @property
    @abstractmethod
    def threads(self) -> List[LabwareThread]:
        raise NotImplementedError

    @abstractmethod
    def get_thread(self, id: str) -> LabwareThread:
        raise NotImplementedError

    @abstractmethod
    def get_thread_by_labware(self, labware_id: str) -> LabwareThread:
        raise NotImplementedError

    @abstractmethod
    def add_thread(self, labware_thread: LabwareThread) -> None:
        raise NotImplementedError

    @abstractmethod
    def create_thread_instance(self, template: ThreadTemplate) -> LabwareThread:
        raise NotImplementedError


class ThreadRegistry(IThreadRegistry):
    def __init__(self, labware_registry: ILabwareRegistry, thread_factory: ThreadFactory) -> None:
        self._threads: Dict[str, LabwareThread] = {}
        self._labware_registry = labware_registry
        self._thread_factory = thread_factory

    @property
    def threads(self) -> List[LabwareThread]:
        return list(self._threads.values())

    def get_thread(self, id: str) -> LabwareThread:
        return self._threads[id]

    def get_thread_by_labware(self, labware_id: str) -> LabwareThread:
        matches = list(filter(lambda thread: thread.labware.id == labware_id, self.threads))
        if len(matches) == 0:
            raise KeyError(f"No thread found for labware {labware_id}")
        if len(matches) > 1:
            raise KeyError(f"Multiple threads found for labware {labware_id}")
        return matches[0]

    def add_thread(self, labware_thread: LabwareThread) -> None:
        if labware_thread.id in self._threads.keys():
            raise KeyError(f"Labware Thread {labware_thread.id} is already defined in the system.  Each labware thread must have a unique id")
        self._threads[labware_thread.id] = labware_thread


    def create_thread_instance(self, template: ThreadTemplate) -> LabwareThread:
        thread = self._thread_factory.create_instance(template)
        thread.initialize_labware()
        for method_template in template.method_resolvers:
            method = method_template.get_instance(self._labware_registry)
            method.assign_thread(template.labware_template, thread)
            thread.append_method_sequence(method)
        self.add_thread(thread)
        return thread


class IThreadManager(IThreadRegistry, ABC):

    @abstractmethod
    async def start_all_threads(self) -> None:
        raise NotImplementedError
    
    @abstractmethod
    def stop_all_threads(self) -> None:
        raise NotImplementedError


class ThreadManager(IThreadManager):
    def __init__(self,
                 thread_registry: IThreadRegistry,
                 system_map: SystemMap,
                 move_handler: MoveHandler) -> None:
        self._thread_registry = thread_registry
        self._system_map = system_map
        self._move_handler = move_handler

    @property
    def threads(self) -> List[LabwareThread]:
        return self._thread_registry.threads

    def get_thread(self, id: str) -> LabwareThread:
        return self._thread_registry.get_thread(id)

    def get_thread_by_labware(self, labware_id: str) -> LabwareThread:
        return self._thread_registry.get_thread_by_labware(labware_id)

    def add_thread(self, labware_thread: LabwareThread) -> None:
        return self._thread_registry.add_thread(labware_thread)

    def create_thread_instance(self, template: ThreadTemplate) -> LabwareThread:
        return self._thread_registry.create_thread_instance(template)

    def has_completed(self) -> bool:
        return all([thread.has_completed() for thread in self._thread_registry.threads])

    @property
    def active_threads(self) -> List[LabwareThread]:
        return [thread for thread in self._thread_registry.threads if not thread.has_completed()]

    @property
    def unstarted_threads(self) -> List[LabwareThread]:
        return [thread for thread in self._thread_registry.threads if thread.status == LabwareThreadStatus.CREATED]

    async def start_all_threads(self) -> None:
        # # self._loop.set_debug(True)
        # self._loop.run_until_complete(self.async_execute())
        await self.async_execute()

    def stop_all_threads(self) -> None:
        for thread in self.threads:
            thread.stop()

    async def async_execute(self) -> None:
        while not self.has_completed():

            for thread in self.active_threads:
                orca_logger.info(f"Thread {thread.name} - {thread.status}")
                if thread in self.unstarted_threads:
                    loop = asyncio.get_running_loop()
                    loop.create_task(thread.start())
                await asyncio.sleep(0.2)
        orca_logger.info("All threads have completed execution.")

class ThreadManagerFactory:
    @staticmethod
    def create_instance(labware_registry: ILabwareRegistry, reservation_manager: IReservationManager, system_map: SystemMap, move_handler: MoveHandler) -> IThreadManager:
        thread_factory = ThreadFactory(labware_registry, move_handler, reservation_manager,system_map)
        thread_registry = ThreadRegistry(labware_registry, thread_factory)
        return ThreadManager(thread_registry, system_map, move_handler)


from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import logging
from orca_driver_interface.driver_interfaces import IDriver
from orca_driver_interface.driver_interfaces import ILabwarePlaceableDriver
from orca.resource_models.labware import Labware

orca_logger = logging.getLogger("orca")

class ResourceUnavailableError(Exception):
    def __init__(self, message: str = "Resource is unavailable.") -> None:
        super().__init__(message)
    
    
class IResource(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError
    

class IInitializableResource(IResource, ABC):
    
    @property
    @abstractmethod
    def is_initialized(self) -> bool:
        raise NotImplementedError
    
    @abstractmethod
    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        raise NotImplementedError
    
    @abstractmethod
    async def initialize(self) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError

class IEquipment(IInitializableResource, ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def is_running(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def execute(self, command: str, options: Dict[str, Any]) -> None:
        raise NotImplementedError
    

class ILabwarePlaceable(ABC):
    @property
    def name(self) -> str:
        raise NotImplementedError
    
    @property
    def labware(self) -> Optional[Labware]:
        raise NotImplementedError
    
    def initialize_labware(self, labware: Labware) -> None:
        # TODO: Make async in future
        # TODO: this will need to be restricted to only initilaizing the labware, probably with a LabwareManager service
        raise NotImplementedError
    
    @abstractmethod
    async def prepare_for_pick(self, labware: Labware) -> None:
        raise NotImplementedError
    
    @abstractmethod
    async def prepare_for_place(self, labware: Labware) -> None:
        raise NotImplementedError

    @abstractmethod
    async def notify_picked(self, labware: Labware) -> None:
        raise NotImplementedError
    
    @abstractmethod
    async def notify_placed(self, labware: Labware) -> None:
        raise NotImplementedError


class Equipment(IEquipment):

    def __init__(self, name: str, driver: IDriver) -> None:
        self._name = name
        self._driver = driver

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def is_initialized(self) -> bool:
        return self._driver.is_initialized
    
    @property
    def is_running(self) -> bool:
        return self._driver.is_running
    
    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        self._driver.set_init_options(init_options)
    
    async def initialize(self) -> None:
        orca_logger.info(f"Initializing...")
        orca_logger.info(f"Name: {self._name}")
        await self._driver.initialize()
        orca_logger.info(f"Initialized")

    async def execute(self, command: str, options: Dict[str, Any]) -> None:
        if command is None:
            raise ValueError(f"{self} - No command to execute")
        orca_logger.info(f"{self} - execute - {command}")
        orca_logger.info(f"{self} - {command} executing...")
        await self._driver.execute(command, options)
        orca_logger.info(f"{self} - {command} executed")

    @property
    def is_connected(self) -> bool:
        return self._driver.is_connected

    async def connect(self) -> None:
        orca_logger.info(f"{self} - Connecting...")
        await self._driver.connect()
        orca_logger.info(f"{self} - Connected")

    async def disconnect(self) -> None:
        orca_logger.info(f"{self} - Disconnecting...")
        await self._driver.disconnect()
        orca_logger.info(f"{self} - Disconnected")



class EquipmentLabwareRegistry:
    def __init__(self) -> None:
        self._stage_labware: Optional[Labware] = None
        self._loaded_labware: List[Labware] = []
    
    @property
    def stage(self) -> Optional[Labware]:
        return self._stage_labware
    
    @property
    def loaded_labware(self) -> List[Labware]:
        return self._loaded_labware

    def initialize_labware(self, labware: Labware) -> None:
        if labware in self._loaded_labware:
            return
        else:
            self._loaded_labware.append(labware)

    def unload_labware_to_stage(self, labware: Labware) -> None:
        if labware not in self._loaded_labware:
            raise ValueError(f"{self} - Labware {labware} not found in loaded labwares")
        if self._stage_labware is not None:
            raise ValueError(f"{self} - Stage already contains labware: {self._stage_labware}.  Unable to unload {labware}")
        self._loaded_labware.remove(labware)
        self._stage_labware = labware

    def load_labware_from_stage(self, labware: Labware) -> None:
        if self._stage_labware != labware:
            raise ValueError(f"{self} - Stage labware {self._stage_labware} does not match labware {labware} to load")
        self._stage_labware = None
        self._loaded_labware.append(labware)

    def set_stage(self, labware: Labware | None) -> None:
        self._stage_labware = labware
 
class LabwareLoadableEquipment(Equipment, ILabwarePlaceable):
    def __init__(self, name: str, driver: ILabwarePlaceableDriver) -> None:
        super().__init__(name, driver)
        self._driver: ILabwarePlaceableDriver = driver
        self._labware_reg = EquipmentLabwareRegistry()

    @property
    def labware(self) -> Optional[Labware]:
        return self._labware_reg.stage
    
    @property
    def loaded_labware(self) -> List[Labware]:
        return self._labware_reg.loaded_labware
    
    def initialize_labware(self, labware: Labware) -> None:
        if labware in self._labware_reg.loaded_labware:
            return
        else:
            self._labware_reg.set_stage(labware)

    async def prepare_for_place(self, labware: Labware) -> None:
        if self._labware_reg.stage is not None:
            raise ValueError(f"{self} - Stage already contains labware: {self._labware_reg.stage}.  Unable to place {labware}")
        orca_logger.info(f"{self} - preparing for place of {labware}")
        await self._driver.prepare_for_place(labware.name, labware.labware_type)

    async def prepare_for_pick(self, labware: Labware) -> None:
        if self._labware_reg.stage == labware:
            return
        else:
            orca_logger.info(f"{self} - preparing for pick of {labware}")
            await self._driver.prepare_for_pick(labware.name, labware.labware_type)
            self._labware_reg.unload_labware_to_stage(labware)

    async def notify_picked(self, labware: Labware) -> None:
        if self._labware_reg.stage != labware:
            raise ValueError(f"{self} - An error has ocurred.  The labware {labware} notified as picked does not match the previously staged labware. "
                              "The wrong labware may have been picked.")
        orca_logger.info(f"{self} - labware {labware} picked from stage")
        self._labware_reg.set_stage(None)
        await self._driver.notify_picked(labware.name, labware.labware_type)
    
    async def notify_placed(self, labware: Labware) -> None:
        if self._labware_reg.stage is not None:
            raise ValueError(f"{self} - An error has ocurred.  The labware {labware} notified as placed was placed with labware {self._labware_reg.stage} already on the stage.  "
                             "A crash may have occurred.")
        self._labware_reg.set_stage(labware)
        orca_logger.info(f"{self} - labware {labware} received on stage")
        await self._driver.notify_placed(labware.name, labware.labware_type)
        self._labware_reg.load_labware_from_stage(labware)

    def __str__(self) -> str:
        return f"Equipment: {self._name}"

    


from abc import ABC, abstractmethod
import asyncio
from orca.resource_models.base_resource import Equipment, IInitializableResource, IResource
from orca.resource_models.resource_pool import EquipmentResourcePool
from orca.resource_models.transporter_resource import TransporterEquipment


from typing import Dict, List

class IResourceRegistryObesrver(ABC):
    @abstractmethod
    def resource_registry_notify(self, event: str, resource: IResource) -> None:
        raise NotImplementedError()


class IResourceRegistry(ABC):

    @property
    @abstractmethod
    def resources(self) -> List[IResource]:
        raise NotImplementedError

    @abstractmethod
    def get_resource(self, name: str) -> IResource:
        raise NotImplementedError

    @abstractmethod
    def add_resource(self, resource: IResource) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def equipments(self) -> List[Equipment]:
        raise NotImplementedError

    @abstractmethod
    def get_equipment(self, name: str) -> Equipment:
        raise NotImplementedError

    @property
    @abstractmethod
    def transporters(self) -> List[TransporterEquipment]:
        raise NotImplementedError

    @abstractmethod
    def get_transporter(self, name: str) -> TransporterEquipment:
        raise NotImplementedError

    @property
    @abstractmethod
    def resource_pools(self) -> List[EquipmentResourcePool]:
        raise NotImplementedError

    @abstractmethod
    def get_resource_pool(self, name: str) -> EquipmentResourcePool:
        raise NotImplementedError

    @abstractmethod
    def add_resource_pool(self, resource_pool: EquipmentResourcePool) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_observer(self, observer: IResourceRegistryObesrver) -> None:
        raise NotImplementedError
    
    @abstractmethod
    async def initialize_all(self) -> None:
        raise NotImplementedError
    
    @abstractmethod
    def clear_resources(self) -> None:
        raise NotImplementedError


class ResourceRegistry(IResourceRegistry):
    def __init__(self) -> None:
        self._resources: Dict[str, IResource] = {}
        self._resource_pools: Dict[str, EquipmentResourcePool] = {}
        self._observers: List[IResourceRegistryObesrver] = []

    @property
    def resources(self) -> List[IResource]:
        return list(self._resources.values())

    @property
    def equipments(self) -> List[Equipment]:
        return [r for r in self._resources.values() if isinstance(r, Equipment)]

    @property
    def transporters(self) -> List[TransporterEquipment]:
        return [r for r in self._resources.values() if isinstance(r, TransporterEquipment)]

    @property
    def resource_pools(self) -> List[EquipmentResourcePool]:
        return list(self._resource_pools.values())

    def get_resource(self, name: str) -> IResource:
        return self._resources[name]

    def get_equipment(self, name: str) -> Equipment:
        resource = self.get_resource(name)
        if not isinstance(resource, Equipment):
            raise ValueError(f"Resource {name} is not an Equipment resource")
        return resource

    def get_transporter(self, name: str) -> TransporterEquipment:
        resource = self.get_resource(name)
        if not isinstance(resource, TransporterEquipment):
            raise ValueError(f"Resource {name} is not an Equipment resource")
        return resource

    def add_resource(self, resource: IResource) -> None:
        name = resource.name
        if name in self._resources.keys():
            raise KeyError(f"Resource {name} is already defined in the system.  Each resource must have a unique name")
        self._resources[name] = resource
        [observer.resource_registry_notify("resource_added", resource) for observer in self._observers]

    def get_resource_pool(self, name: str) -> EquipmentResourcePool:
        return self._resource_pools[name]

    def add_resource_pool(self, resource_pool: EquipmentResourcePool) -> None:
        name = resource_pool.name
        if name in self._resource_pools.keys():
            raise KeyError(f"Resource Pool {name} is already defined in the system.  Each resource pool must have a unique name")
        self._resource_pools[name] = resource_pool

    def add_observer(self, observer: IResourceRegistryObesrver) -> None:
        self._observers.append(observer)

    async def initialize_all(self) -> None:
        await asyncio.gather(*[r.initialize() for r in self._resources.values() if isinstance(r, IInitializableResource)])

    def clear_resources(self) -> None:
        self._resources.clear()
        self._resource_pools.clear()
        self._observers.clear()


"""
Driver management module.
"""
from abc import ABC
import importlib
import importlib.metadata
import importlib.resources
import subprocess
from pydantic import BaseModel, Field
import json
import os
import sys
from typing import Any, Dict, List, Optional
import requests

from orca_driver_interface.driver_interfaces import IDriver


class CommandParamSchema(BaseModel):
    name: str
    type: str = 'string'
    description: str = ''
    required: bool = False


class CommandSchema(BaseModel):
    name: str
    description: str = ''
    params: List[CommandParamSchema] = Field(default_factory=list)


class DriverInfo(BaseModel):
    name: str
    driverPath: str
    driverClass: str
    type: str = 'labwareable'
    description: str = ''
    initParams: List[CommandParamSchema] = Field(default_factory=list)
    commands: List[CommandSchema] = Field(default_factory=list)


class InstalledDriverInfo(DriverInfo):
    packageName: str


class DriverRegistryInfo(BaseModel):
    name: str
    description: str
    url: str


class InstalledDriverRegistry:
    @property
    def installed_drivers(self) -> Dict[str, InstalledDriverInfo]:
        return self._installed_drivers

    def __init__(self) -> None:
        self._prefix = "orca-driver-"
        self.refresh()

    def get_installed_driver_info(self, driver_name: str) -> InstalledDriverInfo:
        return self.installed_drivers[driver_name]

    def is_driver_installed(self, driver_name: str) -> bool:
        return driver_name in self._installed_drivers

    def refresh(self) -> None:
        self._installed_drivers: Dict[str, InstalledDriverInfo] = {}
        self._discover_installed_drivers()

    def _discover_installed_drivers(self) -> None:
        for dist in importlib.metadata.distributions():
            name = dist.metadata["Name"]
            if name is None:
                continue
            if name.startswith(self._prefix):
                driver_name = name.replace(self._prefix, "")
                # get the driver info file
                driver_info_path = importlib.resources.files(f"{driver_name}_driver").joinpath("driver.json")
                driver_info_text = driver_info_path.read_text()
                driver_info: Dict[str, Any] = json.loads(driver_info_text)

                installed_driver_info = InstalledDriverInfo(
                    name=driver_info.get("name", driver_name),
                    description=driver_info.get("description", ""),
                    driverPath=driver_info.get("driverPath", f"{self._prefix}{driver_name}.driver"),
                    driverClass=driver_info.get("driverClass", f"{driver_name.capitalize()}Driver"),
                    packageName=dist.name
                )
                self._installed_drivers[installed_driver_info.name] = installed_driver_info


class AvailableDriverRegistry(ABC):
    def __init__(self, installed_registry: InstalledDriverRegistry) -> None:
        self._registry: Dict[str, DriverRegistryInfo] = {}
        self._installed_registry = installed_registry

    def get_driver_info(self, driver_name: str) -> DriverRegistryInfo:
        self._load_registry()
        if driver_name not in self._registry:
            raise KeyError(f"Driver '{driver_name}' not found in the registry")
        return self._registry[driver_name]

    def get_available_drivers_info(self) -> Dict[str, DriverRegistryInfo]:
        self._load_registry()
        return {
            name: info
            for name, info in self._registry.items()
            if not self._installed_registry.is_driver_installed(name)
        }

    def _load_registry(self) -> None:
        registry_json = self._get_registry_json()
        self._registry = {}
        for driver in registry_json['drivers']:
            driver_info = DriverRegistryInfo(**driver)
            self._registry[driver_info.name] = driver_info

    def _get_registry_json(self) -> Dict[str, Any]:
        raise NotImplementedError


class LocalAvailableDriverRegistry(AvailableDriverRegistry):
    def __init__(self, registry_json_filepath: str, installed_registry: InstalledDriverRegistry) -> None:
        self._registry_filepath = registry_json_filepath
        super().__init__(installed_registry)

    def _get_registry_json(self) -> Dict[str, Any]:
        if os.path.exists(self._registry_filepath):
            with open(self._registry_filepath, 'r') as file:
                return json.load(file)
        else:
            raise FileNotFoundError(f"Registry file '{self._registry_filepath}' not found")


class RemoteAvailableDriverRegistry(AvailableDriverRegistry):
    def __init__(self, registry_json_url: str, installed_registry: InstalledDriverRegistry) -> None:
        self._registry_url = registry_json_url
        super().__init__(installed_registry)

    def _get_registry_json(self) -> Dict[str, Any]:
        response = requests.get(self._registry_url)
        if response.status_code == 200:
            return response.json()
        raise ConnectionError(f"Failed to load driver registry from {self._registry_url}")


class DriverInstaller:
    '''
    Responsible for installing and uninstalling drivers using pip.
    '''
    def __init__(self, installed_registry: InstalledDriverRegistry) -> None:
        self._installed_registry = installed_registry

    def install_driver(self, driver_name: str, driver_repo_url: str) -> None:
        '''
        Installs the driver using pip and registers it in the installed registry.
        '''
        try:
            # Install from the given repository URL (e.g., GitHub)
            install_command = [sys.executable, "-m", "pip", "install", f"git+{driver_repo_url}"]

            # Run the pip install command
            subprocess.check_call(install_command)
            print(f"Driver '{driver_name}' installed successfully.")

            # refresh installed drivers
            self._installed_registry.refresh()

        except subprocess.CalledProcessError as e:
            print(f"Error occurred during installation of driver '{driver_name}': {e}")
            raise

    def uninstall_driver(self, driver_name: str) -> None:
        '''
        Uninstalls the driver using pip.
        '''
        if not self._installed_registry.is_driver_installed(driver_name):
            raise KeyError(f"Driver '{driver_name}' is not installed.")

        driver_info = self._installed_registry.get_installed_driver_info(driver_name)
        package_name = driver_info.packageName

        try:
            uninstall_command = [sys.executable, "-m", "pip", "uninstall", "-y", package_name]
            subprocess.check_call(uninstall_command)
            print(f"Driver '{driver_name}' uninstalled successfully.")

        except subprocess.CalledProcessError as e:
            print(f"Failed to uninstall driver '{driver_name}': {e}")
            raise
        self._installed_registry.refresh()


class DriverLoader:
    '''
    Loads drivers that are installed via pip.
    '''
    def load_driver(self, driver_name: str, installed_registry: InstalledDriverRegistry) -> IDriver:
        try:
            # Retrieve driver info from the installed registry
            if not installed_registry.is_driver_installed(driver_name):
                raise KeyError(f"Driver '{driver_name}' is not installed.")

            driver_info = installed_registry._installed_drivers[driver_name]

            # Dynamically load the driver module and class from driverPath and driverClass
            driver_module = importlib.import_module(driver_info.driverPath)
            driver_class = getattr(driver_module, driver_info.driverClass)

            if not issubclass(driver_class, IDriver):
                raise TypeError(f"Class '{driver_info.driverClass}' does not implement IDriver.")
            args = {"name": driver_name}
            return driver_class(**args)

        except (ImportError, AttributeError, TypeError) as e:
            raise RuntimeError(f"Error loading driver '{driver_name}': {e}") from e


class IDriverManager(ABC):
    def get_driver(self, driver_name: str) -> IDriver:
        raise NotImplementedError

    def install_driver(self, driver_name: str, driver_repo_url: Optional[str]) -> None:
        raise NotImplementedError

    def get_driver_info(self, driver_name: str) -> InstalledDriverInfo:
        raise NotImplementedError

    def get_installed_drivers_info(self) -> Dict[str, InstalledDriverInfo]:
        raise NotImplementedError

    def get_available_drivers_info(self) -> Dict[str, DriverRegistryInfo]:
        raise NotImplementedError

    def uninstall_driver(self, driver_name: str) -> None:
        raise NotImplementedError


class DriverManager(IDriverManager):
    def __init__(self,
                 installed_registry: InstalledDriverRegistry,
                 driver_loader: DriverLoader,
                 driver_installer: DriverInstaller,
                 driver_registry: AvailableDriverRegistry):
        self._installed_registry = installed_registry
        self._driver_loader = driver_loader
        self._driver_installer = driver_installer
        self._driver_registry = driver_registry

    def get_driver(self, driver_name: str) -> IDriver:
        if self._installed_registry.is_driver_installed(driver_name):
            return self._driver_loader.load_driver(driver_name, self._installed_registry)
        else:
            raise KeyError(f"Driver '{driver_name}' is not installed")

    def get_driver_info(self, driver_name: str) -> InstalledDriverInfo:
        return self._installed_registry.installed_drivers[driver_name]

    def get_installed_drivers_info(self) -> Dict[str, InstalledDriverInfo]:
        return self._installed_registry.installed_drivers

    def get_available_drivers_info(self) -> Dict[str, DriverRegistryInfo]:
        return self._driver_registry.get_available_drivers_info()

    def install_driver(self, driver_name: str, driver_repo_url: Optional[str]) -> None:
        if driver_repo_url is None:
            driver_info = self._driver_registry.get_driver_info(driver_name)
            driver_repo_url = driver_info.url

        self._driver_installer.install_driver(driver_name, driver_repo_url)

    def uninstall_driver(self, driver_name: str) -> None:
        self._driver_installer.uninstall_driver(driver_name)



from enum import Enum, auto

class ActionStatus(Enum):
    CREATED = auto()
    RESOLVED = auto()
    AWAITING_LOCATION_RESERVATION = auto()
    AWAITING_CO_THREADS = auto()
    PERFORMING_ACTION = auto()
    AWAITING_MOVE_RESERVATION = auto()
    MOVING = auto()
    COMPLETED = auto()
    ERRORED = auto()

class MethodStatus(Enum):
    CREATED = auto()
    IN_PROGRESS = auto()
    COMPLETED = auto()

class LabwareThreadStatus(Enum):
    UNCREATED = auto()
    CREATED = auto()
    AWAITING_ACTION_RESERVATION = auto()
    AWAITING_MOVE_RESERVATION = auto()
    AWAITING_MOVE_TARGET_AVAILABILITY = auto()
    MOVING = auto()
    AWAITING_CO_THREADS = auto()
    PERFORMING_ACTION = auto()    
    COMPLETED = auto()
    STOPPING = auto()
    STOPPED = auto()






from typing import Any, Dict, Optional
from orca_driver_interface.driver_interfaces import ILabwarePlaceableDriver

class NullPlatePadDriver(ILabwarePlaceableDriver):
    def __init__(self, name: str) -> None:
        self._name = name
        self._is_initialized = False
        self._init_options: Dict[str, Any] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @property
    def is_running(self) -> bool:
        return False

    async def execute(self, command: str, options: Dict[str, Any]) -> None:
        pass

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        self._init_options = init_options

    async def initialize(self) -> None:
        self._is_initialized = True

    @property
    def is_connected(self) -> bool:
        return True

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    async def prepare_for_pick(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        pass

    async def prepare_for_place(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        pass

    async def notify_picked(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        pass

    async def notify_placed(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        pass

    def __str__(self) -> str:
        return f"PlatePad: {self._name}"


from typing import Any, Dict, Optional
from orca_driver_interface.driver_interfaces import ILabwarePlaceableDriver
from orca.driver_management.drivers.null_plate_pad.null_plate_pad import NullPlatePadDriver
from orca.resource_models.base_resource import ILabwarePlaceable, LabwareLoadableEquipment
from orca.resource_models.labware import Labware


class PlatePad(LabwareLoadableEquipment, ILabwarePlaceable):

    def __init__(self, name: str, driver: ILabwarePlaceableDriver = NullPlatePadDriver("Basic Plate Pad")) -> None:
        super().__init__(name, driver)
        self._is_initialized = False
        self._labware: Optional[Labware] = None

    @property
    def labware(self) -> Optional[Labware]:
        return self._labware
    
    def initialize_labware(self, labware: Labware) -> None:
        self._labware = labware

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        return super().set_init_options(init_options)

    async def initialize(self) -> None:
        await self._driver.initialize()
        self._is_initialized = True
    
    async def notify_picked(self, labware: Labware) -> None:
        await self._driver.notify_picked(labware.name, labware.labware_type)
        self._labware = None
    
    async def notify_placed(self, labware: Labware) -> None:
        await self._driver.notify_placed(labware.name, labware.labware_type)
        self._labware = labware
        
    async def prepare_for_pick(self, labware: Labware) -> None:
        if self._labware != labware:
            raise ValueError(f"Labware {labware} not found on plate pad {self}")
        await self._driver.prepare_for_pick(labware.name, labware.labware_type)
    
    async def prepare_for_place(self, labware: Labware) -> None:
        if self._labware != None:
            raise ValueError(f"Trying to place Labware {labware}, but Labware {self._labware} already on plate pad {self}")
        await self._driver.prepare_for_place(labware.name, labware.labware_type)



import os
from typing import Any, Callable, Coroutine, Dict, List, Optional
import logging
import asyncio

import yaml

from orca.driver_management.driver_installer import DriverInstaller, DriverLoader, DriverManager, InstalledDriverRegistry, RemoteAvailableDriverRegistry
from orca.helper import FilepathReconciler
from orca.scripting.scripting import IScriptFactory, IScriptRegistry, NullScriptFactory, ScriptFactory, ScriptRegistry
from orca.system.method_executor import MethodExecutor
from orca.resource_models.base_resource import IInitializableResource
from orca.resource_models.labware import LabwareTemplate
from orca.resource_models.location import Location
from orca.system.move_handler import MoveHandler
from orca.system.registries import LabwareRegistry, TemplateRegistry
from orca.system.reservation_manager import ReservationManager
from orca.system.resource_registry import ResourceRegistry
from orca.system.system import ISystem, SystemInfo
from orca.system.system_map import SystemMap
from orca.system.thread_manager import ThreadManagerFactory
from orca.system.workflow_registry import WorkflowRegistry
from orca.yml_config_builder.configs import SystemConfigModel
from orca.yml_config_builder.template_factories import ConfigToSystemBuilder, MethodTemplateFactory, ThreadTemplateFactory, WorkflowTemplateFactory
from orca.yml_config_builder.resource_factory import IResourceFactory, ResourceFactory



PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
orca_logger = logging.getLogger("orca")
class OrcaCore:

    def __init__(self, 
                 config_filepath: str, 
                 driver_manager: DriverManager,
                 scripting_registry: Optional[IScriptRegistry] = None,
                 resource_factory: Optional[IResourceFactory] = None,
                 deployment_stage: str = "prod"
                 ) -> None:
        self._config_filepath = config_filepath
        with open( self._config_filepath, "r") as f:
            yml_content = f.read()
        self._yml = yaml.load(yml_content, Loader=yaml.FullLoader) # type: ignore
        self._config = SystemConfigModel.model_validate(self._yml)

        # set scripting registry
        if scripting_registry is None:
            absolute_path = os.path.abspath(config_filepath)
            directory_path = os.path.dirname(absolute_path)
            file_reconciler = FilepathReconciler(directory_path)
            # if no scripting is defined, create a null scripting factory
            if self._config.scripting == {}:
                scripting_factory: IScriptFactory = NullScriptFactory()
                scripting_registry = ScriptRegistry(scripting_factory)
            else:
                scripting_factory = ScriptFactory(self._config.scripting, file_reconciler)
                scripting_registry = ScriptRegistry(scripting_factory)
        self._scripting_registry: IScriptRegistry = scripting_registry

        # set resource factory
        if resource_factory is None:
            absolute_path = os.path.abspath(config_filepath)
            directory_path = os.path.dirname(absolute_path)
            filepath_reconciler = FilepathReconciler(directory_path)
            resource_factory = ResourceFactory(driver_manager, filepath_reconciler)
        self._resource_factory: IResourceFactory = resource_factory

        resource_reg = ResourceRegistry()
        system_map = SystemMap(resource_reg)
        template_registry = TemplateRegistry()
        labware_registry = LabwareRegistry()
        reservation_manager = ReservationManager(system_map) 
        move_handler = MoveHandler(reservation_manager, labware_registry, system_map)
        thread_manager = ThreadManagerFactory.create_instance(labware_registry, reservation_manager, system_map, move_handler)
        workflow_registry = WorkflowRegistry(thread_manager, labware_registry, system_map)
        system_info = SystemInfo(self._config.system.name, 
                                 self._config.system.version, 
                                 self._config.system.description, 
                                 dict(self._config.options.__dict__))
        method_template_factory = MethodTemplateFactory(resource_reg, labware_registry)
        thread_template_factory = ThreadTemplateFactory(labware_registry, 
                                                        system_map, 
                                                        template_registry, 
                                                        template_registry, 
                                                        system_map,
                                                        thread_manager,
                                                        self._scripting_registry)
        workflow_template_factory = WorkflowTemplateFactory(thread_template_factory, template_registry)

        self._builder = ConfigToSystemBuilder()
        self._builder.set_config(self._config)
        self._builder.set_system_info(system_info)
        self._builder.set_system_map(system_map)
        self._builder.set_script_registry(self._scripting_registry)
        self._builder.set_resource_registry(resource_reg)
        self._builder.set_template_registry(template_registry)
        self._builder.set_labware_registry(labware_registry)
        self._builder.set_thread_manager(thread_manager)
        self._builder.set_workflow_registry(workflow_registry)
        self._builder.set_resource_factory(resource_factory)
        self._builder.set_method_template_factory(method_template_factory)
        self._builder.set_workflow_template_factory(workflow_template_factory)
        self._method_executor_registry: Dict[str, MethodExecutor] = {}
        self._system = self._builder.get_system(deployment_stage)

    @property
    def system(self) -> ISystem:
        return self._system
    
    @property
    def system_config(self) -> SystemConfigModel:
        return self._config

    async def initialize(self,
                   resource_list: Optional[List[str]] = None,
                   deployment_stage: Optional[str] = None) -> None:
        if deployment_stage is not None:
            self._builder.clear_registries()
            self._system = self._builder.get_system(deployment_stage)
        if resource_list is None:
            await self._system.initialize_all()
        else:
            init_fxns: List[Callable[[], Coroutine[Any, Any, None]]] = []
            for resource_name in resource_list:
                resource = self._system.get_resource(resource_name)
                if isinstance(resource, IInitializableResource):
                    init_fxns.append(resource.initialize)
            await asyncio.gather(*[f() for f in init_fxns])

    def create_workflow_instance(self, 
                                 workflow_name: str, 
                                 deployment_stage: Optional[str] = None,
                                 ) -> str:
        if deployment_stage is not None:
            self._builder.clear_registries()
            self._system = self._builder.get_system(deployment_stage)
        workflow_template = self._system.get_workflow_template(workflow_name)
        workflow = self._system.create_workflow_instance(workflow_template)
        self._system.add_workflow(workflow)
        return workflow.id

    async def run_workflow(self, workflow_id: str) -> None:
        workflow = self._system.get_workflow(workflow_id)
        # TODO:  Add check to make sure system is initailized
        await self.initialize()
        await workflow.start()


    def create_method_instance(self, 
                               method_name: str, 
                               start_map: Dict[str, str], 
                               end_map: Dict[str, str], 
                               deployment_stage: Optional[str] = None,
                               ) -> str:
        if deployment_stage is not None:
            self._builder.clear_registries()
            self._system = self._builder.get_system(deployment_stage)
        try:
            method_template = self._system.get_method_template(method_name)
        except KeyError:
            raise LookupError(f"Method {method_name} is not defined with then System.  Make sure it is included in the config file and the config file loaded correctly.")

        def labware_location_hook(d: Dict[str, str]):
            conversion: Dict[LabwareTemplate, Location] = {}
            for key, value in d.items():
                try:
                    template = self._system.get_labware_template(key)
                except KeyError:
                    raise ValueError(f"Labware {key} is not defined in the system")
                try:
                    location = self._system.get_location(value)
                except KeyError:
                    raise ValueError(f"Location {value} is not defined in the system")
                conversion[template] = location
            return conversion

        start_map_obj: Dict[LabwareTemplate, Location] = labware_location_hook(start_map)
        end_map_obj: Dict[LabwareTemplate, Location] = labware_location_hook(end_map)

        method_template = self._system.get_method_template(method_name)
        executer = MethodExecutor(method_template,
                                 start_map_obj,
                                 end_map_obj,
                                 self._system,
                                 self._system)
        self._method_executor_registry[executer.id] = executer
        return executer.id
    
    async def run_method(self, method_id: str) -> None:
        executer = self._method_executor_registry[method_id]
        # TODO:  Add check to make sure system is initailized
        await self.initialize()
        await executer.start()
        

    def stop(self) -> None:
        self._system.stop_all_threads()


if __name__ == "__main__":
    logging.basicConfig(handlers=[logging.StreamHandler()], level=logging.DEBUG)
    config_file_path =  r"C:\Users\miike\source\repos\orca\orca-core\examples\smc_assay\smc_assay_example.orca.yml"
    workflow_name = "smc-assay"
    available_drivers_registry: str = "https://raw.githubusercontent.com/Cheshire-Labs/orca-extensions/refs/heads/main/drivers.json"
    installed_registry = InstalledDriverRegistry()
    driver_manager = DriverManager(
            installed_registry,
            DriverLoader(), 
            DriverInstaller(installed_registry), 
            RemoteAvailableDriverRegistry(available_drivers_registry, installed_registry))
    orca = OrcaCore(config_file_path, driver_manager)
    workflow_id = orca.create_workflow_instance(workflow_name)
    asyncio.run(orca.run_workflow(workflow_id))


import asyncio
import json
from typing import Any, Dict, List, Optional
from socketio.async_client import AsyncClient # type: ignore[import-untyped]

from orca_driver_interface.driver_interfaces import IDriver, ILabwarePlaceableDriver


class SocketClient:
    def __init__(self, uri: str) -> None:
        self.uri = uri
        # self.cache_ttl = 5  # Cache TTL in seconds
        # self.last_checked_initialized = 0
        # self.last_checked_running = 0
        self.sio = AsyncClient()

    async def connect(self) -> None:
        await self.sio.connect(self.uri)

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    async def disconnect(self):
        await self.sio.disconnect()

    async def send(self, action: str, data: Dict[str, Any]) -> Any:
        await self.sio.emit(action, data)
        response = await self.sio.call(f'{action}_response')
        return response


class RemoteDriverClient(IDriver):
    def __init__(self, name: str):
        self._name: str = name
        self._uri: str | None = None
        self._sio: SocketClient | None = None
        self._init_options: Dict[str, Any] = {}

    @property
    def is_connected(self) -> bool:
        if self._sio is None:
            return False
        return self._sio.is_connected

    async def connect(self) -> None:
        ip = self._init_options["ip"] or "localhost"
        port = self._init_options["port"] or 5001
        self._uri = f"ws://{ip}:{port}"
        self._sio = SocketClient(self._uri)
        await self._sio.connect()

    async def disconnect(self) -> None:
        assert self._sio is not None
        await self._sio.disconnect()

    async def send(self, action: str, data: Dict[str, Any]) -> Any:
        assert self._sio is not None
        return await self._sio.send(action, data)
        
    def _run_async_task(self, action: str, data: Dict[str, Any] = {}) -> Dict[str, Any]:
        try:
            loop = asyncio.get_running_loop()
            future = asyncio.run_coroutine_threadsafe(self.send(action, data), loop)
            return json.loads(future.result())
        except RuntimeError:
            return json.loads(asyncio.run(self.send(action, data)))

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_initialized(self) -> bool:
        dict_ret = self._run_async_task("is_initialized")
        return dict_ret["is_initialized"]

    @property
    def is_running(self) -> bool:
        dict_ret = self._run_async_task("is_running")
        return dict_ret["is_running"]

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        self._run_async_task("set_init_options", init_options)
        
    async def initialize(self) -> None:
        self._run_async_task("initialize")

    async def execute(self, command: str, options: Dict[str, Any]) -> None:
        dict_ret = self._run_async_task("execute", {"command": command, "options": options})


class RemoteLabwarePlaceableDriverClient(RemoteDriverClient, ILabwarePlaceableDriver):
    
    async def prepare_for_pick(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        self._run_async_task("prepare_for_pick", {"labware_name": labware_name, "labware_type": labware_type, "barcode": barcode, "alias": alias})

    async def prepare_for_place(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        self._run_async_task("prepare_for_place", {"labware_name": labware_name, "labware_type": labware_type, "barcode": barcode, "alias": alias})

    async def notify_picked(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        self._run_async_task("notify_picked", {"labware_name": labware_name, "labware_type": labware_type, "barcode": barcode, "alias": alias})

    async def notify_placed(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        self._run_async_task("notify_placed", {"labware_name": labware_name, "labware_type": labware_type, "barcode": barcode, "alias": alias})


class RemoteTransporterDriverClient(RemoteDriverClient):

    async def pick(self, position_name: str, labware_type: str) -> None:
        self._run_async_task(f"pick", {"position_name": position_name, "labware_type": labware_type})

    async def place(self, position_name: str, labware_type: str) -> None:
        self._run_async_task(f"place" , {"position_name": position_name, "labware_type": labware_type})

    def get_taught_positions(self) -> List[str]:
        ret_dict = self._run_async_task("set_init_options")
        return ret_dict["positions"]
    
if __name__ == '__main__':

    s = RemoteDriverClient("test_client")
    s.set_init_options({"ip": "localhost", "port": 5001})
    asyncio.run(s.connect())
    asyncio.run(s.initialize())
    print(s.is_initialized)
    asyncio.run(s.disconnect())


import pytest
import yaml

from orca.yml_config_builder.configs import SystemOptionsConfigModel, ConfigModel, VariablesConfigModel
from orca.yml_config_builder.dynamic_config import DynamicBaseConfigModel
from orca.yml_config_builder.variable_resolution import VariablesRegistry

example_variable_config = r"""
config:
    prod:
        settle-time: 120
        volume: 35.5
        plate-name: plate-1
        recursive: ${opt:stage}
        self-recursion: ${config:prod.plate-name}
    test:
        settle-time: 0
        volume: 40.1
        plate-name: test-plate-1
"""

example_options_config = r"""
opt:
    stage: prod
    lookup: volume
"""

@pytest.fixture
def variables_config() -> VariablesConfigModel:
    yml_config = yaml.safe_load(example_variable_config)
    config = yml_config["config"]
    return VariablesConfigModel.model_validate(config)

@pytest.fixture
def options_config() -> SystemOptionsConfigModel:
    yml_config = yaml.safe_load(example_options_config)
    config = yml_config["opt"]
    return SystemOptionsConfigModel.model_validate(config)

@pytest.fixture
def registry(variables_config, options_config) -> VariablesRegistry:
    registry = VariablesRegistry()
    registry.set_selector_configuration("opt", options_config)
    registry.set_selector_configuration("config", variables_config)
    return registry

class TestVariableConfig:
    def test_variable_config(self, variables_config: VariablesConfigModel):
        assert variables_config.model_extra["test"]["volume"] == 40.1

class TestVariableRegistry:
    def test_no_group(self, registry: VariablesRegistry):
        get_string = "test"
        value = registry.get(get_string)
        assert value == get_string

    def test_selector_parsing(self, registry: VariablesRegistry):
        get_string = r"${config:test.volume}"
        value = registry.get(get_string)
        assert value == 40.1

    def test_nested_parse_config(self, registry: VariablesRegistry):
        get_string = r"${config:${opt:stage}.plate-name}"
        value = registry.get(get_string)
        assert value == "plate-1"

    def test_nested_parse_config2(self, registry: VariablesRegistry):
        get_string = r"${config:test.${opt:lookup}}"
        value = registry.get(get_string)
        assert value == 40.1

    def test_consecutive_subgroups(self, registry: VariablesRegistry):
        get_string = r"${config:${opt:stage}.${opt:lookup}}"
        value = registry.get(get_string)
        assert value == 35.5

    def test_recursive_loopup(self, registry: VariablesRegistry):
        get_string = r"${config:prod.recursive}"
        value = registry.get(get_string)
        assert value == "prod"

    def test_self_recursive_loopup(self, registry: VariablesRegistry):
        get_string = r"${config:prod.self-recursion}"
        value = registry.get(get_string)
        assert value == "plate-1"

class MockConfig(ConfigModel):
    test: str

class DynamicMockClass(DynamicBaseConfigModel[MockConfig]):
    @property
    def test(self) -> str:
        return self.get_dynamic(self._config.test)

class TestConfigWrapper:
    def test_lookup_adapter(self, registry: VariablesRegistry):
        config = MockConfig.model_validate({"test": "${config:test.plate-name}"})
        adapter = DynamicMockClass(config, registry)
        assert adapter.test == "test-plate-1"

    


from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

class ISystemOptionsConfig(ABC):
    pass

class IVariablesConfig(ABC):
    @abstractmethod
    def get_deployment_stages(self) -> List[str]:
        raise NotImplementedError

class IScriptConfig(ABC):
    @property
    @abstractmethod
    def source(self) -> str:
        raise NotImplementedError

class IScriptBaseConfig(ABC):    
    @property
    @abstractmethod
    def base_dir(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def scripts(self) -> Dict[str, IScriptConfig]:
        raise NotImplementedError

class IMethodActionConfig(ABC):    
    @property
    @abstractmethod
    def resource(self) -> Optional[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def command(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def inputs(self) -> List[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def outputs(self) -> Optional[List[str]]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

class IMethodConfig(ABC):    
    @property
    @abstractmethod
    def script(self) -> List[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def on_init(self) -> List[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def actions(self) -> List[Dict[str, IMethodActionConfig]]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

class IThreadStepConfig(ABC):
    @property
    @abstractmethod
    def method(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def spawn(self) -> List[str]:
        raise NotImplementedError

class ILabwareThreadConfig(ABC):
    @property
    @abstractmethod
    def labware(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def start(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def end(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def type(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def scripts(self) -> List[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def steps(self) -> List[Union[str, IThreadStepConfig]]:
        raise NotImplementedError

class ISystemSettingsConfig(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def version(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError

class ILocationConfig(ABC):
    @property
    @abstractmethod
    def teachpoint_name(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

class ILabwareConfig(ABC):
    @property
    @abstractmethod
    def type(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def static(self) -> bool:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

class IResourcePoolConfig(ABC):
    @property
    @abstractmethod
    def type(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def resources(self) -> List[str]:
        raise NotImplementedError

class IResourceConfig(ABC):
    @property
    @abstractmethod
    def type(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def com(self) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def ip(self) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def plate_pad(self) -> Optional[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def sim(self) -> Optional[bool]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def base_type(self) -> Optional[str]:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

class IWorkflowConfig(ABC):

    @property
    @abstractmethod
    def threads(self) -> Dict[str, ILabwareThreadConfig]:
        raise NotImplementedError

class ISystemConfig(ABC):
    @property
    @abstractmethod
    def system(self) -> ISystemSettingsConfig:
        raise NotImplementedError

    @property
    @abstractmethod
    def labwares(self) -> Dict[str, ILabwareConfig]:
        raise NotImplementedError

    @property
    @abstractmethod
    def config(self) -> IVariablesConfig:
        raise NotImplementedError

    @property
    @abstractmethod
    def locations(self) -> Dict[str, ILocationConfig]:
        raise NotImplementedError

    @property
    @abstractmethod
    def resources(self) -> Dict[str, Union[IResourceConfig, IResourcePoolConfig]]:
        raise NotImplementedError

    @property
    @abstractmethod
    def methods(self) -> Dict[str, IMethodConfig]:
        raise NotImplementedError

    @property
    @abstractmethod
    def workflows(self) -> Dict[str, IWorkflowConfig]:
        raise NotImplementedError

    @property
    @abstractmethod
    def scripting(self) -> IScriptBaseConfig | None:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def options(self) -> ISystemOptionsConfig:
        raise NotImplementedError




from typing import Dict, List, Optional, Tuple, Union
import logging
from orca.config_interfaces import ILabwareConfig, ILabwareThreadConfig, IMethodActionConfig, IMethodConfig, IResourceConfig, IResourcePoolConfig, ISystemConfig, IThreadStepConfig, IWorkflowConfig
from orca.resource_models.base_resource import Equipment, LabwareLoadableEquipment
from orca.resource_models.labware import AnyLabwareTemplate, LabwareTemplate
from orca.resource_models.location import Location
from orca.system.workflow_registry import WorkflowRegistry
from orca.scripting.scripting import IScriptRegistry
from orca.system.labware_registry_interfaces import ILabwareTemplateRegistry
from orca.system.thread_manager import IThreadManager
from orca.system.registries import LabwareRegistry
from orca.system.resource_registry import IResourceRegistry
from orca.system.system import System, SystemInfo
from orca.system.system_map import ILocationRegistry, IResourceLocator, SystemMap
from orca.system.registries import TemplateRegistry
from orca.system.template_registry_interfaces import IThreadTemplateRegistry, IMethodTemplateRegistry, IWorkflowTemplateRegistry
from orca.workflow_models.spawn_thread_action import SpawnThreadAction
from orca.workflow_models.labware_thread import IThreadObserver
from orca.yml_config_builder.configs import SystemConfigModel
from orca.yml_config_builder.dynamic_config import DynamicSystemConfigModel
from orca.yml_config_builder.resource_factory import IResourceFactory, ResourcePoolFactory
from orca.resource_models.resource_pool import EquipmentResourcePool
from orca.resource_models.transporter_resource import TransporterEquipment
from orca.workflow_models.workflow_templates import IMethodTemplate, JunctionMethodTemplate, ThreadTemplate, MethodActionTemplate, MethodTemplate, WorkflowTemplate
from orca.yml_config_builder.special_yml_parsing import get_dynamic_yaml_keys, is_dynamic_yaml
from orca.yml_config_builder.variable_resolution import VariablesRegistry


orca_logger = logging.getLogger("orca")
class LabwareConfigToTemplateAdapter:

    def get_template(self, name: str, config: ILabwareConfig) -> LabwareTemplate:
        labware = LabwareTemplate(name, config.type, config.options)
        return labware

class MethodActionConfigToTemplate:
    def __init__(self, resource_reg: IResourceRegistry, labware_temp_reg: ILabwareTemplateRegistry) -> None:
        self._resource_reg: IResourceRegistry = resource_reg
        self._labware_temp_reg: ILabwareTemplateRegistry = labware_temp_reg

    def get_template(self, name: str, config: IMethodActionConfig)  -> MethodActionTemplate:
        # Get the resource name
        resource_name = config.resource
        if resource_name is None:
            resource_name = name
        
        # Get the equipment or resource pool object
        try:
            resource: Equipment | EquipmentResourcePool = self._resource_reg.get_equipment(resource_name)
            
        except KeyError:
            try:
                resource = self._resource_reg.get_resource_pool(resource_name)
            except KeyError:
                raise LookupError(f"The resource name '{resource_name}' in method actions is not recognized as a defined resource or resource pool")
        if not isinstance(resource, Equipment) and not isinstance(resource, EquipmentResourcePool):
            raise TypeError(f"The resource name '{resource_name}' in method actions is not an Equipment or EquipmentResourcePool")

        # Get the input labware templates
        inputs: List[Union[LabwareTemplate, AnyLabwareTemplate]] = []
        for input in config.inputs:
            if input == "$ANY":
                inputs.append(AnyLabwareTemplate())
            else:
                try:
                    labware_template = self._labware_temp_reg.get_labware_template(input)
                except KeyError:
                    raise LookupError(f"The input labware name '{input}' in method actions is not recognized as a defined labware.  Input must be a recognized labware or keyword")
                inputs.append(labware_template)
        # Get the output labware templates
        if config.outputs is None:
            outputs: List[Union[LabwareTemplate, AnyLabwareTemplate]] = inputs
        else:
            outputs = [self._labware_temp_reg.get_labware_template(name) for name in config.outputs]
        
        return MethodActionTemplate(resource,
                                    config.command,
                                    inputs,
                                    outputs, 
                                    config.options)


class MethodTemplateFactory:
    def __init__(self, resource_reg: IResourceRegistry, labware_template_reg: ILabwareTemplateRegistry) -> None:
        self._template_builder = MethodActionConfigToTemplate(resource_reg, labware_template_reg)

    def get_template(self, name: str, config: IMethodConfig) -> MethodTemplate:
        method = MethodTemplate(name, config.options)
        
        for action_item in config.actions:
            for action_name, action_config in action_item.items():
                method_action_template = self._template_builder.get_template(action_name, action_config)
                method.append_action(method_action_template)
        return method    


## Workflow building

class ThreadTemplateFactory:
    def __init__(self,
                labware_temp_reg: ILabwareTemplateRegistry, 
                location_reg: ILocationRegistry, 
                method_temp_reg: IMethodTemplateRegistry, 
                labware_thread_temp_reg: IThreadTemplateRegistry, 
                resource_locator: IResourceLocator,
                thread_manager: IThreadManager,
                script_reg: IScriptRegistry) -> None:
        self._labware_temp_reg: ILabwareTemplateRegistry = labware_temp_reg
        self._location_reg: ILocationRegistry = location_reg
        self._method_temp_reg: IMethodTemplateRegistry = method_temp_reg
        self._thread_temp_reg: IThreadTemplateRegistry = labware_thread_temp_reg
        self._resource_locator: IResourceLocator = resource_locator
        self._thread_manager: IThreadManager = thread_manager
        self._script_reg: IScriptRegistry = script_reg

    def get_template(self, 
                name: str, 
                config: ILabwareThreadConfig) -> ThreadTemplate:
        try:
            labware = self._get_labware_template(config)
            start_location = self._get_location(self._location_reg, config.start)                                            
            end_location = self._get_location(self._location_reg, config.end)
            observers: List[IThreadObserver] = [self._script_reg.get_script(script_name) for script_name in config.scripts]
            labware_thread = ThreadTemplate(labware, 
                                            start_location, 
                                            end_location, 
                                            observers)
        except KeyError as e:
            raise KeyError(f"Error building labware thread {name}") from e
        return labware_thread

    def _get_labware_template(self,
                            config: ILabwareThreadConfig) -> LabwareTemplate:
        try:
            return self._labware_temp_reg.get_labware_template(config.labware)
        except KeyError as e:
            raise KeyError(f"Labware {config.labware} is not in the system labwares") from e

    def _get_location(self, location_registry: ILocationRegistry, location_name: str) -> Location:
        try:
            return location_registry.get_location(location_name)
        except KeyError:
            try:
                return self._resource_locator.get_resource_location(location_name)
            except KeyError:
                 raise KeyError(f"Start/end location {location_name} which is not in the system locations or system resources")
    
    def apply_methods(self, thread_template: ThreadTemplate, config: ILabwareThreadConfig) -> None:
        for step in config.steps:
            if isinstance(step, str):
                method_name = step
                method_resolver = self._get_method_resolver(method_name)
                thread_template.add_method(method_resolver)

            elif isinstance(step, IThreadStepConfig):
                method_resolver = self._get_method_resolver(step.method, step)
                thread_template.add_method(method_resolver)
            else:
                raise ValueError(f"Labware Thread {thread_template.name} has an invalid step type {step}.  Steps must be a string or a ThreadStepConfig")


    def _get_method_resolver(self, method_name: str, step_config: Optional[IThreadStepConfig] = None) -> IMethodTemplate:
        if is_dynamic_yaml(method_name):
            input_index = [int(key) for key in get_dynamic_yaml_keys(method_name)]
            shared_method = JunctionMethodTemplate()
            return shared_method

        try: 
            method = self._method_temp_reg.get_method_template(method_name)
        except KeyError:
            raise LookupError(f"Method {method_name} is not defined.  Method {method_name} must be defined")
        if step_config is not None:
            if step_config.spawn:
                for thread_name in step_config.spawn:
                    thread_template = self._thread_temp_reg.get_labware_thread_template(thread_name)
                    spawn_action = SpawnThreadAction(self._thread_manager, thread_template)
                    method.add_method_observer(spawn_action)
        return method

class WorkflowTemplateFactory:
    def __init__(self,
                 thread_template_factory: ThreadTemplateFactory,
                 thread_temp_reg: IThreadTemplateRegistry) -> None:
        self._thread_temp_reg: IThreadTemplateRegistry = thread_temp_reg
        self._thread_template_factory = thread_template_factory

    def get_template(self, name: str, config: IWorkflowConfig) -> WorkflowTemplate:
        workflow = WorkflowTemplate(name)
        
        thread_templates: List[Tuple[ThreadTemplate, ILabwareThreadConfig]] = []
        # initialize thread templates
        for thread_name, thread_config in config.threads.items():
            thread_template = self._thread_template_factory.get_template(thread_name, thread_config)
            thread_templates.append((thread_template, thread_config))

        # register thread templates
        for thread_template, _ in thread_templates:
            self._thread_temp_reg.add_labware_thread_template(thread_template)

        # appply methods to thread templates
        for thread_template, thread_config in thread_templates:
            self._thread_template_factory.apply_methods(thread_template, thread_config)
            
            if thread_config.type == "start":
                workflow.add_thread(thread_template, is_start=True)
        return workflow

class ScriptsBuilder:
    def __init__(self, config: ISystemConfig, script_reg: IScriptRegistry) -> None:
        self._config: ISystemConfig = config
        self._scripting_registry = script_reg

    def build_scripts(self) -> None:
        if self._config.scripting is None:
            return
        for script_name, script_config in self._config.scripting.scripts.items():
            filepath, class_name = script_config.source.split(":")
            script = self._scripting_registry.create_script(filepath, class_name)
            self._scripting_registry.add_script(script_name, script)


class ResourcesBuilder:
    def __init__(self, config: ISystemConfig, resource_factory: IResourceFactory, resource_reg: IResourceRegistry) -> None:
        self._config: ISystemConfig = config
        self._resource_factory: IResourceFactory = resource_factory
        self._resource_reg: IResourceRegistry = resource_reg

    def build_resources(self) -> None:

        resource_pool_configs: Dict[str, IResourcePoolConfig] = {}

        # build resources from resource defs in config, defer resource pool creation
        for name, resource_config in self._config.resources.items():
            if resource_config.type == "pool" and isinstance(resource_config, IResourcePoolConfig):
                resource_pool_configs[name] = resource_config
                continue
            elif isinstance(resource_config, IResourceConfig):
                self._resource_reg.add_resource(self._resource_factory.create(name, resource_config))
            else:
                raise ValueError(f"Resource {name} has an invalid type {resource_config.type}")
            
        # build resource pools
        for name, resource_config in resource_pool_configs.items():
            pool = ResourcePoolFactory(self._resource_reg).create(name, resource_config)
            self._resource_reg.add_resource_pool(pool)

class LocationsBuilder:
    def __init__(self, config: ISystemConfig, resource_registry: IResourceRegistry, system_map: SystemMap) -> None:
        self._config: ISystemConfig = config
        self._resource_registry: IResourceRegistry = resource_registry
        self._system_map: SystemMap = system_map

    def build_locations(self) -> None:
        # build locations from location defs in config
        for location_name, location_config in self._config.locations.items():
            location = Location(location_config.teachpoint_name)
            location.set_options(location_config.options)  
            self._system_map.add_location(location)
        
        for res in self._resource_registry.resources:
            # skip resources like newtowrk switches, etc that don't have plate pad locations
            if isinstance(res, LabwareLoadableEquipment) \
                and not isinstance(res, EquipmentResourcePool) \
                and not isinstance(res, TransporterEquipment):
                # set resource to each location
                # if the plate-pad is not set in the resource definition, then use the resource name
                resource_config = self._config.resources[res.name]
                if not isinstance(resource_config, IResourceConfig):
                    raise ValueError(f"Resource {res.name} has an invalid type {resource_config.type}")
                if resource_config.plate_pad is not None:
                    location_name = resource_config.plate_pad
                else:
                    location_name = res.name

                try:
                    location = self._system_map.get_location(location_name)
                except KeyError:
                    raise LookupError(f"Location {location_name} referenced in resource {res.name} is not recognized.  Locations must be defined by the transporting resource.")
                location.resource = res





class LabwareTemplatesBuilder:
    def __init__(self, config: ISystemConfig, labware_template_registry: ILabwareTemplateRegistry) -> None:
        self._config = config
        self._adapter = LabwareConfigToTemplateAdapter()
        self._labware_template_registry = labware_template_registry
    
    def build_labware_templates(self) -> None:
        for labware_name, labware_config in self._config.labwares.items():
            labware_template = self._adapter.get_template(labware_name, labware_config)
            self._labware_template_registry.add_labware_template(labware_template)


class MethodTemplatesBuilder:
    def __init__(self, config: ISystemConfig, method_template_factory: MethodTemplateFactory, method_template_registry: IMethodTemplateRegistry) -> None:
        self._config: ISystemConfig = config
        self._method_template_factory: MethodTemplateFactory = method_template_factory
        self._method_template_registry: IMethodTemplateRegistry = method_template_registry

    def build_method_templates(self) -> None:
        for method_name, method_config in self._config.methods.items():
            method_template = self._method_template_factory.get_template(method_name, method_config)
            self._method_template_registry.add_method_template(method_template)
 
class WorkflowTemplatesBuilder:
    def __init__(self, 
                 config: ISystemConfig, 
                 workflow_template_factory: WorkflowTemplateFactory, 
                 workflow_template_registry: IWorkflowTemplateRegistry
                 ) -> None:
        self._config: ISystemConfig = config
        self._workflow_template_factory: WorkflowTemplateFactory = workflow_template_factory
        self._workflow_temp_reg: IWorkflowTemplateRegistry = workflow_template_registry

    def build_workflow_templates(self) -> None:
        for workflow_name, workflow_config in self._config.workflows.items():
            workflow_template = self._workflow_template_factory.get_template(workflow_name, workflow_config)
            self._workflow_temp_reg.add_workflow_template(workflow_template)



class ConfigToSystemBuilder:
    def __init__(self) -> None:
        
        self._config: Optional[ISystemConfig] = None
        self._scripting_registry: Optional[IScriptRegistry] = None
        self._system_info: Optional[SystemInfo] = None 
        self._system_map: Optional[SystemMap] = None
        self._resource_reg: Optional[IResourceRegistry] = None
        self._template_registry: Optional[TemplateRegistry] = None
        self._labware_registry: Optional[LabwareRegistry] = None
        self._thread_manager: Optional[IThreadManager] = None
        self._workflow_registry: Optional[WorkflowRegistry] = None
        self._resource_factory: Optional[IResourceFactory] = None
        self._method_template_factory: Optional[MethodTemplateFactory] = None
        self._workflow_template_factory: Optional[WorkflowTemplateFactory] = None
        self._variable_registry = VariablesRegistry()


    def set_config(self, config: SystemConfigModel) -> None:
        self._variable_registry.set_selector_configuration("labwares", config.labwares)
        self._variable_registry.set_selector_configuration("config", config.config)
        self._config = DynamicSystemConfigModel(config, self._variable_registry)

    def set_system_info(self, system_info: SystemInfo) -> None:
        self._system_info = system_info
    
    def set_system_map(self, system_map: SystemMap) -> None:
        self._system_map = system_map

    def set_resource_registry(self, resource_reg: IResourceRegistry) -> None:
        self._resource_reg = resource_reg

    def set_template_registry(self, template_registry: TemplateRegistry) -> None:
        self._template_registry = template_registry
    
    def set_labware_registry(self, labware_registry: LabwareRegistry) -> None:
        self._labware_registry = labware_registry

    def set_thread_manager(self, thread_manager: IThreadManager) -> None:
        self._thread_manager = thread_manager

    def set_workflow_registry(self, workflow_registry: WorkflowRegistry) -> None:
        self._workflow_registry = workflow_registry

    def set_script_registry(self, script_reg: IScriptRegistry) -> None:
        self._scripting_registry = script_reg
    
    def set_resource_factory(self, resource_factory: IResourceFactory) -> None:
        self._resource_factory = resource_factory

    def set_method_template_factory(self, method_template_factory: MethodTemplateFactory) -> None:
        self._method_template_factory = method_template_factory

    def set_workflow_template_factory(self, workflow_template_factory: WorkflowTemplateFactory) -> None:
        self._workflow_template_factory = workflow_template_factory
    
    def clear_registries(self) -> None:
        if self._resource_reg is not None:
            self._resource_reg.clear_resources()
        if self._template_registry is not None:
            self._template_registry.clear()
        if self._labware_registry is not None:
            self._labware_registry.clear()
        if self._workflow_registry is not None:
            self._workflow_registry.clear()
        if self._scripting_registry is not None:
            self._scripting_registry.clear()
        
    def get_system(self, deployment_stage: str) -> System:
        
        if self._config is None:
            raise ValueError("Config is not set.  You must set the config before building the system")
        if self._scripting_registry is None:
            raise ValueError("Script registry is not set.  You must set the script registry before building the system")
        if self._system_info is None:
            raise ValueError("System info is not set.  You must set the system info before building the system") 
        if self._system_map is None:
            raise ValueError("System map is not set.  You must set the system map before building the system")
        if self._resource_reg is None:
            raise ValueError("Resource registry is not set.  You must set the resource registry before building the system")
        if self._template_registry is None:
            raise ValueError("Template registry is not set.  You must set the template registry before building the system")
        if self._labware_registry is None:
            raise ValueError("Labware registry is not set.  You must set the labware registry before building the system")
        if self._thread_manager is None:
            raise ValueError("Thread manager is not set.  You must set the thread manager before building the system")
        if self._workflow_registry is None:
            raise ValueError("Workflow registry is not set.  You must set the workflow registry before building the system")
        if self._resource_factory is None:
            raise ValueError("Resource factory is not set.  You must set the resource factory before building the system")
        if self._method_template_factory is None:
            raise ValueError("Method template factory is not set.  You must set the method template factory before building the system")
        if self._workflow_template_factory is None:
            raise ValueError("Workflow template factory is not set.  You must set the workflow template factory before building the system")
        
        self._variable_registry.set_selector_configuration("opt", {"stage": deployment_stage})
        orca_logger.info(f"Building system for deployment stage '{deployment_stage}'...")
        
        system = System(self._system_info, 
                        self._system_map, 
                        self._resource_reg, 
                        self._template_registry, 
                        self._labware_registry, 
                        self._thread_manager, 
                        self._workflow_registry)

        
        self._scripting_registry.set_system(system)

        self._scripts_builder = ScriptsBuilder(self._config, self._scripting_registry)
        self._resources_builder = ResourcesBuilder(self._config, self._resource_factory, self._resource_reg)
        self._locations_builder = LocationsBuilder(self._config, self._resource_reg, self._system_map)
        self._labware_templates_builder = LabwareTemplatesBuilder(self._config, self._labware_registry)
        self._workflow_template_builder = WorkflowTemplatesBuilder(self._config, self._workflow_template_factory, self._template_registry)
        self._method_template_builder = MethodTemplatesBuilder(self._config, self._method_template_factory, self._template_registry)

        self._scripts_builder.build_scripts()
        self._resources_builder.build_resources()
        self._locations_builder.build_locations()
        self._labware_templates_builder.build_labware_templates()
        self._method_template_builder.build_method_templates()
        self._workflow_template_builder.build_workflow_templates()
        orca_logger.info("Completed System build")
        return system


from typing import Optional

from typing import Any, Dict
from orca.resource_models.plate_pad import PlatePad
from orca.resource_models.base_resource import ILabwarePlaceable
from orca.resource_models.labware import Labware


from abc import ABC


from typing import List



class IResourceLocationObserver(ABC):
    def location_notify(self, event: str, location: "Location", resource: ILabwarePlaceable) -> None:
        pass

class ILabwareLocationObserver(ABC):
    def notify_labware_location_change(self, event: str, location: "Location", labware: Labware) -> None:
        pass

class Location(ILabwarePlaceable):
    def __init__(self, teachpoint_name: str, resource: Optional[ILabwarePlaceable] = None) -> None:
        self._teachpoint_name = teachpoint_name
        self._resource: ILabwarePlaceable = resource if resource else PlatePad(teachpoint_name)
        self._options: Dict[str, Any] = {}
        self._resource_observers: List[IResourceLocationObserver] = []
        self._labware_observers: List[ILabwareLocationObserver] = []
    
    @property
    def name(self) -> str:
        return self._teachpoint_name
                              
    @property
    def teachpoint_name(self) -> str:
        # TODO: this is redundant to name, these should be refactored into one or differentiated
        return self._teachpoint_name

    @property
    def labware(self) -> Optional[Labware]:
        return self._resource.labware
    
    def initialize_labware(self, labware: Labware) -> None:
        # TODO: this will need to be restricted to only initilaizing the labware
        self._resource.initialize_labware(labware)

    @property
    def resource(self) -> ILabwarePlaceable:
        return self._resource
    
    @resource.setter
    def resource(self, resource: ILabwarePlaceable) -> None:
        self._resource = resource
        for obeserver in self._resource_observers:
            obeserver.location_notify("resource_set", self, resource)
    
    def set_options(self, options: Dict[str, Any]) -> None:
        self._options = options

    async def prepare_for_place(self, labware: Labware) -> None:
        await self._resource.prepare_for_place(labware)

    async def prepare_for_pick(self, labware: Labware) -> None:
        await self._resource.prepare_for_pick(labware)

    async def notify_picked(self, labware: Labware) -> None:
        await self._resource.notify_picked(labware)
        for observer in self._labware_observers:
            observer.notify_labware_location_change("picked", self, labware)
    
    async def notify_placed(self, labware: Labware) -> None:
        await self._resource.notify_placed(labware)
        for observer in self._labware_observers:
            observer.notify_labware_location_change("placed", self, labware)

    def __str__(self) -> str:
        return f"Location: {self._teachpoint_name}"
    
    def add_observer(self, observer: IResourceLocationObserver | ILabwareLocationObserver) -> None:
        if isinstance(observer, ILabwareLocationObserver):
            if observer in self._labware_observers:
                return
            self._labware_observers.append(observer)
        elif isinstance(observer, IResourceLocationObserver):
            if observer in self._resource_observers:
                return
            self._resource_observers.append(observer)
        else:
            raise NotImplementedError(f"Observer type {type(observer)} not supported")
    



import uuid
from orca.resource_models.base_resource import Equipment
from orca.resource_models.labware import AnyLabwareTemplate, Labware, LabwareTemplate
from orca.resource_models.location import Location
from orca.resource_models.resource_pool import EquipmentResourcePool
from orca.system.labware_registry_interfaces import ILabwareRegistry

from typing import Any, Dict, List, Set, Union

from orca.system.reservation_manager import IReservationManager
from orca.system.system_map import IResourceLocator, SystemMap
from orca.workflow_models.action import AssignedLabwareManager, IActionObserver, LocationAction, LocationActionCollectionReservationRequest
from orca.workflow_models.status_enums import ActionStatus


class DynamicResourceAction:
    def __init__(self,
                 labware_registry: ILabwareRegistry,
                 resource: EquipmentResourcePool | List[Equipment] | Equipment,
                 command: str,
                 expected_input_templates: List[Union[LabwareTemplate, AnyLabwareTemplate]],
                 expected_output_templates: List[Union[LabwareTemplate, AnyLabwareTemplate]],
                 options: Dict[str, Any] = {}) -> None:
        if isinstance(resource, EquipmentResourcePool):
            self._resource_pool: EquipmentResourcePool = resource
        elif isinstance(resource, list):
            self._resource_pool = EquipmentResourcePool(f"Generated Resource Pool - {uuid.uuid4()}", resource)
        elif isinstance(resource, Equipment):
            self._resource_pool = EquipmentResourcePool(f"Generated Resource Pool - {uuid.uuid4()}", [resource])
        self._command: str = command
        self._options: Dict[str, Any] = options
        self._assigned_labware_manager = AssignedLabwareManager(labware_registry, 
                                                                                  expected_input_templates, 
                                                                                  expected_output_templates)
        self._location_action: LocationAction | None = None
        self._observers: List[IActionObserver] = []

    @property
    def command(self) -> str:
        return self._command
    
    @property
    def options(self) -> Dict[str, Any]:
        return self._options

    @property
    def resource_pool(self) -> EquipmentResourcePool:
        return self._resource_pool
    
    @property
    def expected_input_templates(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._assigned_labware_manager.expected_input_templates
    
    @property
    def expected_output_templates(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._assigned_labware_manager.expected_output_templates

    @property
    def expected_inputs(self) -> List[Labware]:
        return self._assigned_labware_manager.expected_inputs
    
    @property
    def expected_outputs(self) -> List[Labware]:
        return self._assigned_labware_manager.expected_outputs
       
    def assign_input(self, template_slot: LabwareTemplate, input: Labware):
        self._assigned_labware_manager.assign_input(template_slot, input)
    
    @property
    def status(self) -> ActionStatus:
        if self._location_action is None:
            return ActionStatus.AWAITING_LOCATION_RESERVATION
        return self._location_action.status

    def add_observer(self, observer: IActionObserver) -> None:
        if observer not in self._observers:
            self._observers.append(observer)
        if self._location_action is not None:
            self._location_action.add_observer(observer)

    async def resolve_action(self, reference_point: Location, reservation_manager: IReservationManager, system_map: SystemMap) -> LocationAction:
        if self._location_action is not None:
            return self._location_action
        reservation_request = LocationActionCollectionReservationRequest(self._get_potential_location_actions(system_map))
        self._location_action = await reservation_request.reserve_location(reservation_manager, reference_point, system_map)
        for observer in self._observers:
            self._location_action.add_observer(observer)
        self._location_action._status = ActionStatus.RESOLVED
        return self._location_action

    def _get_potential_location_actions(self, resource_locator: IResourceLocator) -> List[LocationAction]:
        potential_locations: Set[Location] = set()
        for resource in self._resource_pool.resources:
            potential_location = resource_locator.get_resource_location(resource.name)
            potential_locations.add(potential_location)

        location_actions: List[LocationAction] = []
        for location in potential_locations:
            location_action = LocationAction(location, 
                                             self._command, 
                                             self._assigned_labware_manager,
                                             self._options)
            location_actions.append(location_action)
        return location_actions
    
    
    def __str__(self) -> str:
        return f"Resource Action Pool: {self._resource_pool.name} - {self._command}"



            





import asyncio
import logging
from typing import Dict, List, Optional
import uuid
import networkx as nx # type: ignore
from orca.resource_models.labware import Labware
from orca.resource_models.location import ILabwareLocationObserver, Location
from orca.system.system_map import ILocationRegistry

orca_logger = logging.getLogger("orca")

class ILabwareLocationManager:
    def is_occupied(self, location: Location) -> bool:
        raise NotImplementedError

class IAvailabilityManager:
    def is_location_available(self, location: Location) -> bool:
        raise NotImplementedError
    
    async def await_available(self, location: Location) -> None:
        raise NotImplementedError


class IReservationManager:
    def can_reserve(self, location_name: str) -> bool:
        raise NotImplementedError
    
    def release_reservation(self, location_name: str) -> None:
        raise NotImplementedError

    def request_reservation(self, location_name: str, reservation: "LocationReservation") -> None:
        raise NotImplementedError


class LocationReservation: 
    def __init__(self, requested_location: Location, labware: Labware | None = None) -> None:
        self._id = str(uuid.uuid4())
        self._labware = labware
        self._requested_location: Location = requested_location
        self._reserved_location: Optional[Location] = None
        self._reservation_manager: IReservationManager | None = None
        self.completed: asyncio.Event  = asyncio.Event()
        self.deadlocked: asyncio.Event = asyncio.Event()
        self.cancelled: asyncio.Event = asyncio.Event()

    @property
    def id(self) -> str:
        return self._id
    
    @property
    def labware(self) -> Labware | None:
        return self._labware

    def set_location(self, location: Location) -> None:
        if self.cancelled.is_set():
            raise ValueError("Reservation has been cancelled")
        self._reserved_location = location
        self.completed.set()
    
    @property
    def requested_location(self) -> Location:
        return self._requested_location

    @property
    def reserved_location(self) -> Location:
        if self._reserved_location is None:
            raise ValueError("Location not yet reserved")
        return self._reserved_location

    def request_reservation(self, reservation_manager: IReservationManager) -> None:
        self._reservation_manager = reservation_manager
        reservation_manager.request_reservation(self._requested_location.name, self) 

    def release_reservation(self) -> None:
        if self._reservation_manager is None:
            raise ValueError("Reservation manager not set")
        self._reservation_manager.release_reservation(self.reserved_location.name)
    
class ReservationManager(IReservationManager, IAvailabilityManager, ILabwareLocationObserver):
    def __init__(self, location_reg: ILocationRegistry) -> None:
        self._location_reg = location_reg
        self._location_reservations: Dict[str, LocationReservation] = {}
        self._location_queues: Dict[str, List[LocationReservation]] = {}
        self._deadlock_detector = DeadlockDetector(location_reg, self._location_reservations, self._location_queues)

    def request_reservation(self, location_name: str, request: LocationReservation) -> None:
        self._location_reg.get_location(location_name).add_observer(self)
        self._add_to_waitlist(location_name, request)
        self._process_next_request(location_name)

    def _reserve(self, location_name: str, request: LocationReservation) -> None:
        self._location_reservations[location_name] = request
        request.set_location(self._location_reg.get_location(location_name))
        request.completed.set()
        orca_logger.info(f"Thread {request.labware} - Reservation {request.id} granted for {location_name}")

    def _process_next_request(self, location_name: str) -> None:
        queue = self._location_queues.setdefault(location_name, [])
        if len(queue) == 0:
            return
        if self.can_reserve(location_name):
            request = queue.pop(0)
            self._reserve(location_name, request)
        else:
            if self._deadlock_detector.is_deadlocked():
                self._handle_deadlock(queue.pop(0))
            

    def _add_to_waitlist(self, location_name: str, request: LocationReservation) -> None:
        queue = self._location_queues.setdefault(location_name, [])
        if request not in queue:
            queue.append(request)
            orca_logger.info(f"Thread {request.labware} - Reservation {request.id} waiting for availability at {location_name}")
    
    def can_reserve(self, location_name: str) -> bool:
        return location_name not in self._location_reservations.keys() and self._location_reg.get_location(location_name).labware is None

    def release_reservation(self, location_name: str) -> None:
        if location_name in self._location_reservations.keys():
            reservation = self._location_reservations[location_name]
            orca_logger.info(f"Releasing reservation {reservation.id} for {location_name}")
            del self._location_reservations[location_name]
            self._process_next_request(location_name)
            

    def notify_labware_location_change(self, event: str, location: Location, labware: Labware) -> None:
        if event == "picked":
            if self.can_reserve(location.name):
                self._process_next_request(location.name)

    def _handle_deadlock(self, request: LocationReservation):
        orca_logger.info("Deadlock detected")
        request.deadlocked.set()
        request.completed.set()


class DeadlockDetector:
    def __init__(self, location_registry: ILocationRegistry, location_reservations: Dict[str, LocationReservation], reservation_queues: Dict[str, List[LocationReservation]]) -> None:
        self._location_registry = location_registry
        self._location_reservations = location_reservations
        self._reservation_queues = reservation_queues
    
    def is_deadlocked(self) -> bool:
        graph = DeadlockGraph()
        for location_name, queue in self._reservation_queues.items():
            for request in queue:
                requestor = request.labware
                holder = self._location_registry.get_location(location_name).labware
                if holder is None or requestor is None:
                    continue
                graph._add_edge(requestor, holder)

        return graph.is_deadlocked()

class DeadlockGraph:
    def __init__(self) -> None:
        self._graph: nx.DiGraph = nx.DiGraph()
        
    def _add_edge(self, requester: Labware, holder: Labware) -> None:
        if not self._graph.has_node(requester): # type: ignore
            self._graph.add_node(requester) # type: ignore
        if not self._graph.has_node(holder): # type: ignore
            self._graph.add_node(holder) # type: ignore
        self._graph.add_edge(requester, holder) # type: ignore
        
    def reset(self) -> None:
        self._graph.clear() # type: ignore

    def is_deadlocked(self) -> bool:
        try:
            cycle = nx.find_cycle(self._graph, orientation='original') # type: ignore
            return True
        except nx.NetworkXNoCycle:
            return False







import logging
import time
from typing import Any, Dict, Optional

from orca_driver_interface.driver_interfaces import IDriver

orca_logger = logging.getLogger("orca")

class SimulationBaseDriver(IDriver):
    def __init__(self, name: str, mocking_type: Optional[str] = None, sim_time: float = 0.2):
        self._name: str = name
        self._mocking_type = mocking_type
        self._init_options: Dict[str, Any] = {}
        self._is_initialized: bool = False
        self._is_running: bool = False
        self._connected: bool = False
        self._sim_time = sim_time

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @property
    def is_running(self) -> bool:
        return self._is_running

    @property
    def is_connected(self) -> bool:
        return self._connected

    async def connect(self) -> None:
        self._connected = True

    async def disconnect(self) -> None:
        self._connected = False

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        self._init_options = init_options

    async def initialize(self) -> None:
        self._sleep()
        self._is_initialized = True

    async def execute(self, command: str, options: Dict[str, Any]) -> None:
        orca_logger.info(f"{self._name} executing command: {command}")
        if len(options.keys()) > 0:
            orca_logger.info(f"Options: {options}")
        self._sleep()
        orca_logger.info(f"{self._name} executed command: {command}")

    def _sleep(self) -> None:
        self._is_running = True
        time.sleep(self._sim_time)
        self._is_running = False



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Set, Union
from orca.resource_models.base_resource import Equipment
from orca.resource_models.location import Location

from orca.resource_models.labware import AnyLabwareTemplate, LabwareTemplate
from orca.resource_models.resource_pool import EquipmentResourcePool

from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.workflow_models.dynamic_resource_action import DynamicResourceAction
from orca.workflow_models.labware_thread import IMethodObserver, IThreadObserver, Method


class MethodActionTemplate:
    def __init__(self, resource: Equipment | EquipmentResourcePool,
                 command: str,
                 inputs: List[Union[LabwareTemplate, AnyLabwareTemplate]],
                 outputs: Optional[List[Union[LabwareTemplate, AnyLabwareTemplate]]] = None,
                 options: Optional[Dict[str, Any]] = None):
        if isinstance(resource, Equipment):
            self._resource_pool: EquipmentResourcePool = EquipmentResourcePool(resource.name, [resource])
        else:
            self._resource_pool = resource
        self._command = command
        self._options: Dict[str, Any] = {} if options is None else options
        self._inputs: List[Union[LabwareTemplate, AnyLabwareTemplate]] = inputs
        self._outputs: List[Union[LabwareTemplate, AnyLabwareTemplate]] = outputs if outputs is not None else []

    @property
    def resource_pool(self) -> EquipmentResourcePool:
        return self._resource_pool

    @property
    def inputs(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._inputs
    
    @property
    def outputs(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._outputs

    @property
    def command(self) -> str:
        return self._command

    @property
    def options(self) -> Dict[str, Any]:
        return self._options

class IMethodTemplate(ABC):
    @abstractmethod
    def get_instance(self, labware_reg: ILabwareRegistry) -> Method:
        raise NotImplementedError("IMethodTemplate must implement get_instance method")


class MethodActionFactory:

    def __init__(self, template: MethodActionTemplate, labware_reg: ILabwareRegistry) -> None:
        self._template: MethodActionTemplate = template
        self._labware_reg = labware_reg

    def create_instance(self) -> DynamicResourceAction:

        # TODO: since refactoring, this MethodActionTemplate and DynamicResourceAction are very similar

        instance = DynamicResourceAction(self._labware_reg,
                                        self._template.resource_pool,
                                        self._template.command,
                                        self._template.inputs,
                                        self._template.outputs,
                                        self._template.options,
                                        )
        return instance

class MethodTemplate(IMethodTemplate):

    def __init__(self, name: str, options: Dict[str, Any] = {}):
        self._name = name
        self._actions: List[MethodActionTemplate] = []
        self._method_observers: List[IMethodObserver] = []
        self._options = options


    @property
    def name(self) -> str:
        return self._name

    @property
    def actions(self) -> List[MethodActionTemplate]:
        return self._actions
    
    @property
    def inputs(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        inputs: Set[Union[LabwareTemplate, AnyLabwareTemplate]] = set()
        for action in self._actions:
            inputs.update(action.inputs)
        return list(inputs)
    
    @property
    def outputs(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        outputs: Set[Union[LabwareTemplate, AnyLabwareTemplate]] = set()
        for action in self._actions:
            outputs.update(action.outputs)
        return list(outputs)

    def append_action(self, action: MethodActionTemplate):
        self._actions.append(action)

    def add_method_observer(self, observer: IMethodObserver) -> None:
        self._method_observers.append(observer)

    def get_instance(self, labware_reg: ILabwareRegistry) -> Method:
        method = Method(self.name)
        for action_template in self.actions:
            factory = MethodActionFactory(action_template, labware_reg)
            action = factory.create_instance()
            method.append_step(action)
            for o in self._method_observers:
                method.add_observer(o)

        return method
    
class JunctionMethodTemplate(IMethodTemplate):
    def __init__(self) -> None:
        self._method: Method | None = None
    
    def set_method(self, method: Method) -> None:
        self._method = method

    def get_instance(self, labware_reg: ILabwareRegistry) -> Method:
        if self._method is None:
            raise NotImplementedError("Method has not been set")
        return self._method
    
class ThreadTemplate:

    def __init__(self, labware_template: LabwareTemplate, start: Location, end: Location, observers: List[IThreadObserver] = []) -> None:
        self._labware_template: LabwareTemplate = labware_template
        self._start: Location = start
        self._end: Location = end
        self._methods: List[IMethodTemplate] = []
        self._wrapped_method: Method | None = None
        self._observers: List[IThreadObserver] = observers

    @property
    def name(self) -> str:
        return self._labware_template.name

    @property
    def labware_template(self) -> LabwareTemplate:
        return self._labware_template
    
    @property
    def start_location(self) -> Location:
        return self._start
    
    @property
    def end_location(self) -> Location:
        return self._end
    
    @property
    def method_resolvers(self) -> List[IMethodTemplate]:
        return self._methods
    
    @property
    def observers(self) -> List[IThreadObserver]:
        return self._observers
    
    def set_wrapped_method(self, wrapped_method: Method) -> None:
        self._wrapped_method
        for m in self._methods:
            if isinstance(m, JunctionMethodTemplate):
                m.set_method(wrapped_method)  
    
    def add_method(self, method: IMethodTemplate) -> None:
        self._methods.append(method)


class WorkflowTemplate:
    
    def __init__(self, name: str) -> None:
        self._name = name
        self._start_threads: Dict[str, ThreadTemplate] = {}
        self._threads: Dict[str, ThreadTemplate] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def thread_templates(self) -> List[ThreadTemplate]:
        return list(self._threads.values())

    @property
    def start_thread_templates(self) -> List[ThreadTemplate]:
        return list(self._start_threads.values())
    
    def add_thread(self, thread: ThreadTemplate, is_start: bool = False) -> None:
        self._threads[thread.name] = thread
        if is_start:
            self._start_threads[thread.name] = thread



import asyncio

from orca.resource_models.labware import Labware
from orca.system.system_map import SystemMap


class TestSystemGraph:
    
    def test_get_shortest_path(self, system_map: SystemMap):
        expected_path = ["stacker1", "loc3", "ham1"]
        path = system_map.get_all_shortest_available_paths("stacker1", "ham1")

        assert path[0] == expected_path

    def test_no_path_through_in_use_plate(self, system_map: SystemMap):
        loc3 = system_map.get_location("loc3")
        blocking_labware = Labware("plate", labware_type="mock_labware")
        asyncio.run(loc3.prepare_for_place(blocking_labware))
        asyncio.run(loc3.notify_placed(blocking_labware))
        has_path = system_map.has_available_route("stacker1", "ham1")
        assert not has_path
    



import os

# TODO: This should be applied on paths in the configuration file loader instead of later in the process
class FilepathReconciler:
    def __init__(self, configuration_file_directory: str, base_dir: str = '') -> None:
        self._config_file_dir = configuration_file_directory
        self._base_dir = base_dir        

    def set_base_dir(self, base_dir: str) -> None:
        self._base_dir = base_dir

    def reconcile_filepath(self, filepath: str) -> str:
        # check the current working directory from where Orca was called to find the file
        filepath = os.path.join(self._base_dir, filepath)
        cwd_filepath =  os.path.join(os.getcwd(), filepath)
        
        if os.path.isfile(cwd_filepath):
            return cwd_filepath
        
        # if it doesn't exist in the current working directory, check the directory from which the configuration was loaded
        config_root_filepath = os.path.join(self._config_file_dir, filepath)
        if os.path.isfile(config_root_filepath):
            return config_root_filepath
        
        # else throw an error of file not found
        raise FileNotFoundError(f"File does not exist at either {cwd_filepath} nor {config_root_filepath}")
        


from typing import List
from orca.resource_models.base_resource import Equipment
from orca.resource_models.location import Location

    
class EquipmentResourcePool:

    def __init__(self, name: str, resources: List[Equipment] = []):
        self._name = name
        self._resources: List[Equipment] = resources

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def resources(self) -> List[Equipment]:
        return self._resources

    def add_resource(self, resource: Equipment) -> None:
        self._resources.append(resource)



from __future__ import annotations

from typing import Any, Dict
import uuid

class LabwareTemplate:
    def __init__(self, name: str, type: str, options: Dict[str, Any] = {}) -> None:
        self._name = name
        self._labware_type = type
        self._options = options
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def labware_type(self) -> str:
        return self._labware_type
    
    @labware_type.setter
    def labware_type(self, value: str) -> None:
        self._labware_type = value

    def create_instance(self) -> Labware:
        return Labware(self.name, self.labware_type)

class AnyLabwareTemplate:
    """A class that represents any labware template"""

    @property
    def name(self) -> str:
        return "$any"

    def create_instance(self) -> AnyLabware:
        return AnyLabware()
    
    def __str__(self) -> str:
        return "$AnyLabwareTemplate"


class AnyLabware:
    """A class that represents any labware"""

    @property
    def name(self) -> str:
        return "$any"

    def __str__(self) -> str:
        return "$AnyLabware"


class Labware:
    
    def __init__(self, name: str, labware_type: str) -> None:
        self._id = str(uuid.uuid4())
        self._init_options: Dict[str, Any] = {}
        self._name = name
        self._labware_type = labware_type
    
    @property
    def name(self) -> str:
        return f"{self._name}"

    @property   
    def id(self) -> str:
        return self._id
    
    @property
    def labware_type(self) -> str:
        return self._labware_type

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        self._init_options = init_options

    def __str__(self) -> str:
        return f"{self._name}"


import time
import threading
import asyncio
import logging
from logging import Handler, LogRecord
from typing import Any, Dict, List

from typing import Any, Dict, List

from fastapi import FastAPI, HTTPException
import uvicorn

from orca.cli.socketio_mount import socketio_mount
from orca.cli.orca_api import OrcaApi
from orca.logger.socketio_logger_handler import SocketIOHandler
from orca.resource_models.labware import AnyLabwareTemplate, LabwareTemplate

orca_logger = logging.getLogger("orca")
app = FastAPI()
sio = socketio_mount(app)
socketio_handler = SocketIOHandler(sio)
if not any(isinstance(h, type(socketio_handler)) for h in orca_logger.handlers):
    orca_logger.addHandler(socketio_handler)
    orca_logger.setLevel(logging.DEBUG)

orca_api: OrcaApi = OrcaApi()


@sio.on("connect", namespace="/logging")  # type: ignore
async def handle_connect(sid, environ) -> None:
    print(f"Client connected: {sid}")

@sio.on("disconnect", namespace="/logging")  # type: ignore
async def handle_disconnect(sid) -> None:
    print(f"Client disconnected: {sid}")




# REST API endpoints
@app.post("/load")
async def load(data: Dict[str, Any]) -> Dict[str, str]:
    config_file = data.get("configFile")
    orca_logger.info(f"Configuration loaded from {config_file}")
    if config_file is None:
        raise HTTPException(status_code=400, detail="Config file is required.")
    orca_api.load(config_file)
    return {"message": "Configuration loaded successfully."}



@app.post("/init")
async def init(data: Dict[str, Any]):
    config_file = data.get("configFile")
    resource_list = data.get("resourceList", [])
    stage = data.get("stage", None)
    orca_api.init(config_file, resource_list, stage)
    return {"message": "Initialization complete."}


@app.post("/run_workflow")
async def run_workflow(data: Dict[str, Any]) -> Dict[str, Any]:
    workflow_name = data.get("workflowName")
    if workflow_name is None:
        raise HTTPException(status_code=400, detail="Workflow name is required.")
    config_file = data.get("configFile", None)
    stage = data.get("stage", None)
    workflow_id = orca_api.run_workflow(
        workflow_name, config_file, stage
    )
    orca_logger.info(f"Workflow {workflow_name} started with ID {workflow_id}")
    return {"workflowId": workflow_id}


@app.post("/run_method")
async def run_method(data: Dict[str, Any]) -> Dict[str, Any]:
    method_name = data.get("methodName")
    if method_name is None:
        raise HTTPException(status_code=400, detail="Method name is required.")
    start_map = data.get("startMap", {})
    end_map = data.get("endMap", {})
    config_file = data.get("configFile", None)
    stage = data.get("stage", None)
    method_id = orca_api.run_method(
        method_name,
        start_map,
        end_map,
        config_file,
        stage
    )
    orca_logger.info(f"Method {method_name} started with ID {method_id}")
    return {"methodId": method_id}


@app.get("/get_deployment_stages")
async def get_deployment_stages() -> Dict[str, Any]:
    deployment_stages = orca_api.get_deployment_stages()
    return {"deploymentStages": deployment_stages}

@app.get("/get_workflow_recipes")
async def get_workflow_recipes() -> Dict[str, Any]:
    recipes = orca_api.get_workflow_recipes()
    dict_recipes = {}
    for name, r in recipes.items():
        thread_recipes = {}
        for tr in r.thread_templates:
            thread_recipes[tr.name] = {
                "name": tr.name,
                "startLocation": tr.start_location.name,
                "endLocation": tr.end_location.name,
                "labwareTemplate": tr.labware_template.name,
            }
        dict_recipes[name] = {
            "name": r.name,
            "threadRecipes": thread_recipes,
        }

    return {"workflowRecipes": dict_recipes}


@app.get("/test")
async def test() -> Dict[str, str]:
    return {"status": "route reachable"}


@app.get("/get_method_recipes")
async def get_method_recipes() -> Dict[str, Any]:
    recipes = orca_api.get_method_recipes()
    dict_recipes = {}
    for name, r in recipes.items():
        dict_recipes[name] = {
            "name": r.name,
            "inputs": [labware.name for labware in r.inputs],
            "outputs": [labware.name for labware in r.outputs],
        }
    return {"methodRecipes": dict_recipes}


@app.get("/get_running_workflows")
async def get_running_workflows() -> Dict[str, Any]:
    running_workflows = orca_api.get_running_workflows()
    return {"workflows": running_workflows}


@app.get("/get_running_methods")
async def get_running_methods() -> Dict[str, Any]:
    running_methods = orca_api.get_running_methods()
    return {"methods": running_methods}


@app.get("/get_method_recipe_input_labwares/{method_name}")
def get_method_recipe_input_labwares(method_name: str) -> Dict[str, Any]:
    method_recipe = orca_api.get_method_recipes()[method_name]
    labware_inputs: List[str] = []
    any_count: int = 0
    for labware in method_recipe.inputs:
        if isinstance(labware, AnyLabwareTemplate):
            labware_inputs.append(f"$ANY_{any_count}")
            any_count += 1
        elif isinstance(labware, LabwareTemplate):
            labware_inputs.append(labware.name)
        else:
            raise TypeError(
                f"Labware {labware} is not a recognized labware template type"
            )
    return {"inputLabwares": labware_inputs}


@app.get("/get_method_recipe_output_labwares/{method_name}")
def get_method_recipe_output_labwares(method_name: str) -> Dict[str, Any]:
    method_recipe = orca_api.get_method_recipes()[method_name]
    labware_outputs: List[str] = []
    any_count: int = 0
    for labware in method_recipe.outputs:
        if isinstance(labware, AnyLabwareTemplate):
            labware_outputs.append(f"$ANY_{any_count}")
            any_count += 1
        elif isinstance(labware, LabwareTemplate):
            labware_outputs.append(labware.name)
        else:
            raise TypeError(
                f"Labware {labware} is not a recognized labware template type"
            )
    return {"outputLabwares": labware_outputs}


@app.get("/get_labware_recipes")
def get_labware_recipes() -> Dict[str, Any]:
    recipes = orca_api.get_labware_recipes()
    return {"labwareRecipes": recipes}


@app.get("/get_locations")
def get_locations() -> Dict[str, Any]:
    locations = orca_api.get_locations()
    return {"locations": locations}


@app.get("/get_resources")
def get_resources() -> Dict[str, Any]:
    resources = orca_api.get_resources()
    return {"resources": resources}


@app.get("/get_equipments")
def get_equipments() -> Dict[str, Any]:
    equipments = orca_api.get_equipments()
    return {"equipments": equipments}


@app.get("/get_transporters")
def get_transporters() -> Dict[str, Any]:
    transporters = orca_api.get_transporters()
    return {"transporters": transporters}


@app.post("/stop")
async def stop() -> Dict[str, str]:
    orca_api.stop()
    return {"message": "Orca stopped."}


@app.get("/get_installed_drivers_info")
async def get_installed_drivers_info() -> Dict[str, Any]:
    driver_info = orca_api.get_installed_drivers_info()
    drivers = {name: info.model_dump_json() for name, info in driver_info.items()}
    return {"installedDriversInfo": drivers}


@app.get("/get_available_drivers_info")
async def get_available_drivers_info() -> Dict[str, Any]:
    driver_info = orca_api.get_available_drivers_info()
    drivers = {name: info.model_dump_json() for name, info in driver_info.items()}
    return {"availableDriversInfo": drivers}


@app.post("/install_driver/")
async def install_driver(data: Dict[str, Any]) -> Dict[str, str]:
    driver_name = data.get("driverName", None)
    driver_repo_url = data.get("driverRepoUrl", None)
    if driver_name is None:
        raise HTTPException(status_code=400, detail="Driver name is required.")
    orca_api.install_driver(driver_name, driver_repo_url)
    return {"message": f"Driver '{driver_name}' installed successfully."}


@app.post("/uninstall_driver")
async def uninstall_driver(data: Dict[str, Any]) -> Dict[str, str]:
    driver_name = data.get("driverName", None)
    if driver_name is None:
        raise HTTPException(status_code=400, detail="Driver name is required.")
    orca_api.uninstall_driver(driver_name)
    return {"message": f"Driver '{driver_name}' uninstalled successfully."}


@app.get("/shutdown")
async def shutdown() -> Dict[str, str]:
    """API route to shut down the server."""
    try:
        response = {"message": "Server shutdown: success"}

        orca_logger.info("Shutdown request received, shutting down Orca server")
        loop = asyncio.get_running_loop()
        loop.call_later(1, uvicorn_server.stop)

        return response
    except Exception as e:
        orca_logger.error(f"Error sending shutdown response: {e}")
        return {"message": "Server shutdown: failed"}


# Uvicorn server instance for graceful shutdown
class UvicornServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    def __init__(self, config: uvicorn.Config):
        super().__init__(config)
        self.thread = None

    def start(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

        # Wait for server to start
        while not self.started:
            time.sleep(1e-3)

        return self.thread

    def stop(self):
        self.should_exit = True
        if self.thread:
            self.thread.should_abort_immediately = True  # type: ignore
            self.thread = None


uvicorn_server = UvicornServer(
    config=uvicorn.Config(
        app=app, host="127.0.0.1", port=5000, log_level="debug", loop="asyncio"
    )
)

if __name__ == "__main__":
    uvicorn_server.run()



from typing import Dict
import uuid
from orca.resource_models.labware import LabwareTemplate
from orca.resource_models.location import Location
from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.system.thread_manager import IThreadManager
from orca.workflow_models.spawn_thread_action import SpawnThreadAction
from orca.workflow_models.workflow_templates import JunctionMethodTemplate, MethodTemplate, ThreadTemplate

class MethodExecutor:
    def __init__(self, 
                 template: MethodTemplate, 
                 labware_start_mapping: Dict[LabwareTemplate, Location], 
                 labware_end_mapping: Dict[LabwareTemplate, Location],
                 thread_manager: IThreadManager,
                 labware_registry: ILabwareRegistry) -> None: 
        self._id: str = str(uuid.uuid4())
        self._method_template = template

        # TODO:  Obviously an atrocious fix here to remove any spawn thread actions from the method observers
        # TODO: needs to be replaced after refactoring the registry to make a distinction between 
        # stand alone method template definitonss and method templates defined as part of workflows
        replacement_observers = [o for o in self._method_template._method_observers if not isinstance(o, SpawnThreadAction)]
        self._method_template._method_observers = replacement_observers

        self._start_mapping = labware_start_mapping
        self._end_mapping = labware_end_mapping
        self._thread_manager = thread_manager
        self._labware_registry = labware_registry
        self._validate_labware_location_mappings()
        self._create_labware_threads()

    def _validate_labware_location_mappings(self) -> None:
        # simple check that the AnyLabware wildcard is satisfied
        if len(self._start_mapping) != len(self._method_template.inputs):
            raise ValueError(f"Number of labware in the start_map does not match the number of expected inputs")
    
        if len(self._end_mapping) != len(self._method_template.outputs):
            raise ValueError(f"Number of labware in the end_map does not match the number of expected outputs")
        
        # validate that each concrete labware template is in the maps
        for labware_template in self._method_template.inputs:
            if isinstance(labware_template, LabwareTemplate) and labware_template not in self._start_mapping.keys():
                raise ValueError(f"Labware {labware_template.name} is expected as an input but its starting location is not in the start_map")
        
        for labware_template in self._method_template.outputs:
            if isinstance(labware_template, LabwareTemplate) and labware_template not in self._end_mapping.keys():
                raise ValueError(f"Labware {labware_template.name} is expected as an output but its ending location is not in the end_map")

    def _create_labware_threads(self) -> None:
        # TODO: mappings won't work here for labwares that end or start within a method action
        method = self._method_template.get_instance(self._labware_registry)
        
        for idx, labware_template in enumerate(self._start_mapping.keys()):
            thread_template = ThreadTemplate(labware_template, 
                                             self._start_mapping[labware_template],
                                             self._end_mapping[labware_template])
            thread_template.add_method(JunctionMethodTemplate())
            thread_template.set_wrapped_method(method)
            thread = self._thread_manager.create_thread_instance(thread_template)
            
    @property
    def id(self) -> str:
        return self._id
    
    async def start(self):
        await self._thread_manager.start_all_threads()






import re

def is_dynamic_yaml(value: str) -> bool:
    return re.search(r"\$\{([^}]+)\}", value) is not None

def get_dynamic_yaml_keys(value: str) -> list:
    return re.findall(r"\$\{([^}]+)\}", value)



from typing import List, cast
from abc import ABC, abstractmethod
from orca.config_interfaces import IResourceConfig, IResourcePoolConfig
from orca.driver_management.driver_installer import IDriverManager
from orca.driver_management.driver_socket_client import RemoteLabwarePlaceableDriverClient
from orca.driver_management.drivers.simulation_base.simulation_base import SimulationBaseDriver
from orca.driver_management.drivers.simulation_robotic_arm.simulation_robotic_arm import SimulationRoboticArmDriver
from orca.driver_management.drivers.simulation_labware_placeable.simulation_labware_placeable import SimulationLabwarePlaceableDriver
from orca.helper import FilepathReconciler
from orca.resource_models.base_resource import Equipment, IEquipment, LabwareLoadableEquipment

from orca.resource_models.resource_pool import EquipmentResourcePool
from orca.resource_models.transporter_resource import TransporterEquipment
from orca.system.resource_registry import IResourceRegistry
from orca_driver_interface.driver_interfaces import IDriver, ILabwarePlaceableDriver
from orca_driver_interface.transporter_interfaces import ITransporterDriver

class IResourceFactory(ABC):
    @abstractmethod
    def create(self, resource_name: str, resource_config: IResourceConfig) -> IEquipment:
        raise NotImplementedError
    
class ResourceFactory(IResourceFactory):
    def __init__(self, driver_manager: IDriverManager, filepath_reconciler: FilepathReconciler) -> None:
        self._driver_manager = driver_manager
        self._filepath_reconciler = filepath_reconciler


        # self._resource_map: Dict[str, Callable[[str, IResourceConfig], IEquipment]] = {
        #     'mock-labware-loadable': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Mock Labware Loadable", "Mock Labware Loadable")),
        #     'mock-robot': lambda name, config: TransporterEquipment(name, SimulationRoboticArmDriver(name, filepath_reconciler, "Mock Robot" )),
        #     'ml-star': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Hamilton MLSTAR", "Hamilton MLSTAR")),
        #     'venus': lambda name, config: LabwareLoadableEquipment(name, cast(ILabwarePlaceableDriver, self._driver_manager.get_driver("venus"))),
        #     'test': lambda name, config: LabwareLoadableEquipment(name, cast(ILabwarePlaceableDriver, self._driver_manager.get_driver('test-driver'))),
        #     'acell': lambda name, config: TransporterEquipment(name, SimulationRoboticArmDriver(name, filepath_reconciler, "ACell")),
        #     'ddr': lambda name, config: TransporterEquipment(name, SimulationRoboticArmDriver(name, filepath_reconciler, "DDR")),
        #     'translator': lambda name, config: TransporterEquipment(name, SimulationRoboticArmDriver(name, filepath_reconciler, "Translator")),
        #     'cwash': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("CWash", "CWash")),
        #     'mantis': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Mantis", "Mantis")),
        #     'analytic-jena': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Analytic Jena", "Analytic Jena")),
        #     'tapestation-4200': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Tapestation 4200", "Tapestation 4200")),
        #     'biotek': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Biotek", "Biotek")),
        #     'bravo': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Bravo", "Bravo")),
        #     'plateloc': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Plateloc", "Plateloc")),
        #     'vspin': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("VSpin", "VSpin")),
        #     'agilent-hotel': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Agilent Hotel", "Agilent Hotel")),
        #     'smc-pro': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("SMC Pro", "SMC Pro")),
        #     'vstack': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("VStack", "VStack")),
        #     'shaker': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Shaker", "Shaker")),
        #     'waste': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Waste", "Waste")),
        #     'delidder': lambda name, config: LabwareLoadableEquipment(name, SimulationLabwarePlaceableDriver("Delidder", "Delidder")),
        #     'serial-switch': lambda name, config: Equipment(name, SimulationBaseDriver("Serial Switch", "Serial Switch")),
        #     'switch': lambda name, config: Equipment(name, SimulationBaseDriver("Switch", "Switch"))
        # }

    def create_venus(self, resource_name: str, resource_config: IResourceConfig) -> IEquipment:
        if resource_config.sim:
            return LabwareLoadableEquipment(resource_name, SimulationLabwarePlaceableDriver("Venus", "Venus"))
        else:
            return LabwareLoadableEquipment(resource_name, RemoteLabwarePlaceableDriverClient("venus"))

    def create(self, resource_name: str, resource_config: IResourceConfig) -> IEquipment:
        driver_name = resource_config.type.lower()
        
        # Determine the driver base type
        if resource_config.sim:
            driver_base_type = resource_config.base_type or self._get_driver_base_type(driver_name)
            driver = self._get_similation_driver(resource_name, driver_base_type)
        else:
            driver_info = self._driver_manager.get_driver_info(driver_name)
            driver_base_type = driver_info.type
            driver = self._driver_manager.get_driver(driver_name)
        
        # Create the appropriate equipment based on the driver base type
        if driver_base_type == "transporter":
            equipment: IEquipment = TransporterEquipment(resource_name, cast(ITransporterDriver, driver))
        elif driver_base_type == "non-labware":
            equipment = Equipment(resource_name, driver)
        else:
            equipment = LabwareLoadableEquipment(resource_name, cast(ILabwarePlaceableDriver, driver))

        equipment.set_init_options(resource_config.options)
        return equipment

    def _get_driver_base_type(self, driver_name: str) -> str:
        try:
            driver_info = self._driver_manager.get_driver_info(driver_name)
            return driver_info.type
        except:
            return "labwareable"
    
    def _get_similation_driver(self, name: str, res_type: str) -> IDriver:
        if res_type == 'non-labware':
            return SimulationBaseDriver(name,res_type)
        elif res_type == 'transporter':
            return SimulationRoboticArmDriver(name, self._filepath_reconciler, res_type)
        else:
            return SimulationLabwarePlaceableDriver(name, res_type)


    
class ResourcePoolFactory:
    def __init__(self, resource_reg: IResourceRegistry) -> None:
        self._resource_reg = resource_reg

    def create(self, pool_name: str, pool_config: IResourcePoolConfig) -> EquipmentResourcePool:
        res_type = pool_config.type
        if res_type != 'pool':
            raise ValueError(f"Resource pool {pool_name} type set as {res_type} instead of 'pool'")
        resources: List[Equipment] = []
        for res_name in pool_config.resources:
            try:
                res = self._resource_reg.get_resource(res_name)
            except KeyError:
                raise KeyError(f"Resource {res_name} from resource pool {pool_name} not found in system")
            if not isinstance(res, Equipment):
                raise ValueError(f"Resource {res_name} from resource pool {pool_name} is not a valid equipment resource")
            resources.append(res)
            
        return EquipmentResourcePool(pool_name, resources)    


from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import Dict, List
import uuid
from orca.resource_models.location import Location
from orca.resource_models.resource_pool import EquipmentResourcePool
from orca.resource_models.transporter_resource import TransporterEquipment
from orca.resource_models.base_resource import Equipment, IResource

from orca.resource_models.labware import Labware, LabwareTemplate
from orca.system.labware_registry_interfaces import ILabwareRegistry, ILabwareTemplateRegistry
from orca.system.workflow_registry import IWorkflowRegistry
from orca.system.resource_registry import IResourceRegistry, IResourceRegistryObesrver
from orca.system.system_map import ILocationRegistry, SystemMap
from orca.system.registries import LabwareRegistry, TemplateRegistry
from orca.system.template_registry_interfaces import IThreadTemplateRegistry, IMethodTemplateRegistry, IWorkflowTemplateRegistry
from orca.workflow_models.labware_thread import LabwareThread, Method
from orca.workflow_models.workflow import Workflow
from orca.workflow_models.workflow_templates import ThreadTemplate, MethodTemplate, WorkflowTemplate
from orca.system.thread_manager import IThreadManager

class ISystemInfo(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def version(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError

class SystemInfo(ISystemInfo):
    def __init__(self, name: str, version: str, description: str, model_extra: Dict[str, str]) -> None:
        self._id = str(uuid.uuid4())
        self._name = name
        self._version = version
        self._description = description
        self._model_extra = model_extra
    
    @property
    def id(self) -> str:
        return self._id
    
    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def description(self) -> str:
        return self._description

    @property
    def model_extra(self) -> Dict[str, str]:
        return self._model_extra
    

class ISystem(ISystemInfo,
              IResourceRegistry, 
              ILabwareRegistry, 
              ILabwareTemplateRegistry, 
              IWorkflowTemplateRegistry, 
              IMethodTemplateRegistry, 
              IThreadTemplateRegistry, 
              ILocationRegistry, 
              IWorkflowRegistry, 
              IThreadManager,
              ABC):

    @property
    @abstractmethod
    def system_map(self) -> SystemMap:
        raise NotImplementedError

class System(ISystem):
    def __init__(self, 
                 info: SystemInfo, 
                 system_map: SystemMap, 
                 resource_registry: IResourceRegistry, 
                 template_registry: TemplateRegistry, 
                 labware_registry: LabwareRegistry, 
                 thread_manager: IThreadManager,
                 workflow_registry: IWorkflowRegistry) -> None:
        self._info = info
        self._resources = resource_registry
        self._system_map = system_map
        self._templates = template_registry
        self._labwares = labware_registry
        self._workflow_registry = workflow_registry
        self._thread_manager = thread_manager

    @property
    def id(self) -> str:
        return self._info.id

        
    @property
    def name(self) -> str:
        return self._info.name

    @property
    def version(self) -> str:
        return self._info.version

    @property
    def description(self) -> str:
        return self._info.description
    
    @property
    def system_map(self) -> SystemMap:
        return self._system_map
    
    @property
    def locations(self) -> List[Location]:
        return self._system_map.locations

    @property
    def labwares(self) -> List[Labware]:
        return self._labwares.labwares
    
    @property
    def resources(self) -> List[IResource]:
        return self._resources.resources

    @property
    def equipments(self) -> List[Equipment]:
        return self._resources.equipments
    
    @property
    def transporters(self) -> List[TransporterEquipment]:
        return self._resources.transporters
    
    @property
    def resource_pools(self) -> List[EquipmentResourcePool]:
        return self._resources.resource_pools
    
    @property
    def threads(self) -> List[LabwareThread]:
        return self._thread_manager.threads

    def get_resource(self, name: str) -> IResource:
        return self._resources.get_resource(name)

    def get_equipment(self, name: str) -> Equipment:
        return self._resources.get_equipment(name)

    def get_transporter(self, name: str) -> TransporterEquipment:
        return self._resources.get_transporter(name)

    def get_resource_pool(self, name: str) -> EquipmentResourcePool:
        return self._resources.get_resource_pool(name)

    def get_location(self, name: str) -> Location:
        return self._system_map.get_location(name)

    def get_labware(self, name: str) -> Labware:
        return self._labwares.get_labware(name)

    def add_resource(self, resource: IResource) -> None:
        self._resources.add_resource(resource)

    def add_location(self, location: Location) -> None:
        self._system_map.add_location(location)

    def add_resource_pool(self, resource_pool: EquipmentResourcePool) -> None:
        self._resources.add_resource_pool(resource_pool)

    def add_labware(self, labware: Labware) -> None:
        self._labwares.add_labware(labware)

    def get_labware_template(self, name: str) -> LabwareTemplate:
        return self._labwares.get_labware_template(name)
    
    def add_labware_template(self, labware: LabwareTemplate) -> None:
        self._labwares.add_labware_template(labware)
    
    def get_labware_thread_template(self, name: str) -> ThreadTemplate:
        return self._templates.get_labware_thread_template(name)
    
    def add_labware_thread_template(self, thread: ThreadTemplate) -> None:
        self._templates.add_labware_thread_template(thread)

    def get_method_templates(self) -> MappingProxyType[str, MethodTemplate]:
        return self._templates.get_method_templates()

    def get_method_template(self, name: str) -> MethodTemplate:
        return self._templates.get_method_template(name)
    
    def add_method_template(self, method: MethodTemplate) -> None:
        self._templates.add_method_template(method)

    def get_workflow_templates(self) -> MappingProxyType[str, WorkflowTemplate]:
        return self._templates.get_workflow_templates()

    def get_workflow_template(self, name: str) -> WorkflowTemplate:
        return self._templates.get_workflow_template(name)
    
    def add_workflow_template(self, workflow: WorkflowTemplate) -> None:
        self._templates.add_workflow_template(workflow)

    def get_workflow(self, id: str) -> Workflow:
        return self._workflow_registry.get_workflow(id)
    
    def add_workflow(self, workflow: Workflow) -> None:
        self._workflow_registry.add_workflow(workflow)

    def get_thread(self, id: str) -> LabwareThread:
        return self._thread_manager.get_thread(id)
    
    def get_thread_by_labware(self, labware_id: str) -> LabwareThread:
        return self._thread_manager.get_thread_by_labware(labware_id)

    def add_thread(self, labware_thread: LabwareThread) -> None:
        self._thread_manager.add_thread(labware_thread)

    def get_method(self, id: str) -> Method:
        return self._workflow_registry.get_method(id)
    
    def add_method(self, method: Method) -> None:
        self._workflow_registry.add_method(method)

    def add_observer(self, observer: IResourceRegistryObesrver) -> None:
        return self._resources.add_observer(observer)
    
    def create_method_instance(self, template: MethodTemplate) -> Method:
        return self._workflow_registry.create_method_instance(template)
    
    def create_thread_instance(self, template: ThreadTemplate) -> LabwareThread:
        return self._thread_manager.create_thread_instance(template)
    
    def create_workflow_instance(self, template: WorkflowTemplate) -> Workflow:
        return self._workflow_registry.create_workflow_instance(template)
    
    async def start_all_threads(self) -> None:
        return await self._thread_manager.start_all_threads()

    def stop_all_threads(self) -> None:
        return self._thread_manager.stop_all_threads()
    
    async def initialize_all(self) -> None:
        return await self._resources.initialize_all()

    def clear_resources(self) -> None:
        return self._resources.clear_resources()


from __future__ import annotations
from abc import ABC, abstractmethod
import itertools
from typing import Any, Dict, List, Optional, Set, Tuple
from orca.resource_models.base_resource import IResource, ILabwarePlaceable
from orca.resource_models.labware import Labware
from orca.resource_models.location import IResourceLocationObserver, Location
import networkx as nx # type: ignore
import matplotlib.pyplot as plt

from orca.resource_models.plate_pad import PlatePad
from orca.resource_models.transporter_resource import TransporterEquipment
from orca.system.resource_registry import IResourceRegistry
from orca.system.resource_registry import IResourceRegistryObesrver

class IResourceLocator(ABC):
    def get_resource_location(self, resource_name: str) -> Location:
        raise NotImplementedError

class _NetworkXHandler:
    
    def __init__(self, graph: Optional[nx.DiGraph] = None) -> None: # type: ignore
        if graph is None:
            graph = nx.DiGraph()
        self._graph: nx.DiGraph = graph
    
    def add_node(self, name: str, location: Location) -> None:
        self._graph.add_node(name, location=location) # type: ignore
    
    def add_edge(self, start: str, end: str, transporter: TransporterEquipment, weight: float = 1.0) -> None:
        self._graph.add_edge(start, end, weight=weight, transporter=transporter) # type: ignore

    def has_path(self, source: str, target: str) -> bool:
        return nx.has_path(self._graph, source, target) # type: ignore

    def get_nodes(self) -> Dict[str, Dict[str, Any]]:
        return dict(self._graph.nodes.items())
    
    def get_node_data(self, name: str) -> Dict[str, Any]:
        return self._graph.nodes[name]
    
    def get_shortest_path(self, source: str, target: str) -> List[str]:
        path: List[str] = nx.shortest_path(self._graph, source, target, weight='weight') # type: ignore
        return path

    def get_all_shortest_paths(self, source: str, target: str) -> List[List[str]]:
        return list(nx.all_shortest_paths(self._graph, source, target, weight='weight')) # type: ignore

    def get_all_simple_paths(self, source: str, target: str) -> List[List[str]]:
        return list(nx.all_simple_paths(self._graph, source, target)) # type: ignore

    def get_subgraph(self, nodes: List[str]) -> _NetworkXHandler:
        return _NetworkXHandler(nx.subgraph(self._graph, nodes)) # type: ignore
    
    def get_path_graph(self, path: List[str]) -> _NetworkXHandler:
        return _NetworkXHandler(nx.path_graph(self._graph, path)) # type: ignore
    
    def get_all_edges(self) -> List[Tuple[str, str, Dict[str, Any]]]:
        """Returns all edges with data as a list of tuples (source, target, data)"""
        return list(self._graph.edges(data=True))

    def get_edge_data(self, source: str, target: str) -> Dict[str, Any]:
        return self._graph.edges[source, target]
    
    def get_distance(self, source: str, target: str) -> float:
        return nx.shortest_path_length(self._graph, source, target, weight='weight') # type: ignore

    def draw(self) -> None:
        pos = nx.spring_layout(self._graph) # type: ignore
        nx.draw(self._graph, pos=pos, with_labels=True) # type: ignore
        plt.show() # type: ignore

    def __getitem__(self, key: str) -> Dict[str, Any]:
        return self._graph.nodes[key]


class ILocationRegistry(ABC):
    @property
    @abstractmethod
    def locations(self) -> List[Location]:
        raise NotImplementedError
    
    @abstractmethod
    def get_location(self, name: str) -> Location:
        raise NotImplementedError

    @abstractmethod
    def add_location(self, location: Location):
        raise NotImplementedError
    

class SystemMap(ILocationRegistry, IResourceLocator, IResourceLocationObserver, IResourceRegistryObesrver):

    def __init__(self, resource_registry: IResourceRegistry) -> None:
        self._graph: _NetworkXHandler = _NetworkXHandler()
        self._equipment_map: Dict[str, Location] = {}
        self._resource_registry = resource_registry
        for transporter in self._resource_registry.transporters:
            self.add_transporter(transporter)
        self._resource_registry.add_observer(self)

    @property
    def locations(self) -> List[Location]:
        return [nodedata["location"] for _, nodedata in self._graph.get_nodes().items()]

    def get_location(self, name: str) -> Location:
        return self._graph.get_node_data(name)["location"]

    def add_location(self, location: Location) -> None:
        self._graph.add_node(location.teachpoint_name, location=location)
        if isinstance(location.resource, ILabwarePlaceable):
            self._equipment_map[location.resource.name] = location
        location.add_observer(self)

    def get_resource_location(self, resource_name: str) -> Location:
        try:
            return self._equipment_map[resource_name]
        except KeyError:
            resource_name = resource_name.replace("-", "_")
            try:
                return self._equipment_map[resource_name]
            except KeyError:
                raise ValueError(f"Resource {resource_name} does not exist")
        
    def get_distance(self, source: str, target: str) -> float:
        return self._graph.get_distance(source, target)
    
    def get_transporter_between(self, source: str, target: str) -> TransporterEquipment:
        return self._graph.get_edge_data(source, target)["transporter"]

    def add_edge(self, start: str, end: str, transporter: TransporterEquipment, weight: float = 5.0) -> None:
        if start not in self._graph.get_nodes():
            raise ValueError(f"Node {start} does not exist")
        if end not in self._graph.get_nodes():
            raise ValueError(f"Node {end} does not exist")
        self._graph.add_edge(start, end, transporter=transporter, weight=weight) 

    def set_edge_weight(self, start: str, end: str, weight: float) -> None:
        self._graph[start][end]['weight'] = weight

    def has_available_route(self, source: str, target: str) -> bool:
        available_graph = self._get_available_graph([source])
        return available_graph.has_path(source, target)
    
    def has_any_route(self, source: str, target: str) -> bool:
        return self._graph.has_path(source, target)
        
    def get_all_shortest_available_paths(self, source: str, target: str) -> List[List[str]]:
        available_graph = self._get_available_graph([source])
        return available_graph.get_all_shortest_paths(source, target)
    
    def get_all_shortest_any_paths(self, source: str, target: str) -> List[List[str]]:
        return self._graph.get_all_shortest_paths(source, target)

    def get_shortest_paths_to_deadlock_resolution(self, source: str) -> List[List[str]]:
        paths = []
        for name, data in self._graph.get_nodes().items():
            if isinstance(data["location"].resource, PlatePad) and name != source:
                paths.extend(self.get_all_shortest_any_paths(source, name)) #type: ignore
            
        return paths
    
    def _get_blocking_locations(self, source: str, target: str) -> List[Location]:
        # TODO: add input validations of source and target entered
        unique_stop = {stop for path in self.get_all_shortest_any_paths(source, target) for stop in path}
        blocking_locs: Set[Location] = set()
        for location_name in unique_stop:
            if location_name == source:
                continue
            location: Location = self._graph.get_node_data(location_name)["location"]
            if location is not None:
                blocking_locs.add(location)
        return list(blocking_locs)

    def _get_blocking_transporter(self, labware: Labware, source: str, target: str) -> List[TransporterEquipment]:
        blocking_transporters: Set[TransporterEquipment] = set()
        for path in self.get_all_shortest_any_paths(source, target):
            for i in range(len(path) - 1):
                edge = self._graph.get_edge_data(path[i], path[i + 1])
                transporter: TransporterEquipment = edge["transporter"]
                if transporter.labware is not None:
                    blocking_transporters.add(transporter)
        return list(blocking_transporters)
    
    def draw(self) -> None:
        self._graph.draw()
        
    def add_transporter(self, transporter: TransporterEquipment) -> None:
        taught_locations = transporter.get_taught_positions()
        # add teachpoints as locations if they don't exist and connect them as an edge
        for edge in itertools.combinations(taught_locations, 2):
            try:
                self.get_location(edge[0])
            except KeyError:
                self.add_location(Location(edge[0]))
            try:
                self.get_location(edge[1])
            except KeyError:
                self.add_location(Location(edge[1]))
            self.add_edge(edge[0], edge[1], transporter)
            self.add_edge(edge[1], edge[0], transporter)

    def resource_registry_notify(self, event: str, resource: IResource) -> None:
        if event == "resource_added":
            if isinstance(resource, TransporterEquipment):
                self.add_transporter(resource)

    def location_notify(self, event: str, location: Location, resource: ILabwarePlaceable) -> None:
        if event == "resource_set":
            if isinstance(resource, ILabwarePlaceable):
                self._equipment_map[resource.name] = location

    def _get_available_graph(self, include_nodes: List[str] = []) -> _NetworkXHandler:
        subgraph = self._graph.get_subgraph([name for name, _ in self._get_available_locations(include_nodes).items()])
        available_edges: List[Tuple[str, str, Dict[str, Any]]] = []
        for edge in subgraph.get_all_edges():
            source, target, data = edge
            transporter: TransporterEquipment = data["transporter"]
            if transporter.labware is None:
                available_edges.append(edge)
        available_graph = _NetworkXHandler()
        for name, nodedata in subgraph.get_nodes().items(): 
            available_graph.add_node(name, nodedata["location"])
        for source, target, data in available_edges:
            available_graph.add_edge(source, target, data["transporter"]) 
        return available_graph


    def _get_available_locations(self, include_nodes: List[str] = []) -> Dict[str, Dict[str, Any]]:
        nodes = {}
        for node, nodedata in self._graph.get_nodes().items():
            location: Location = nodedata["location"]
            if node in include_nodes or location.labware is None:
                nodes[node] = nodedata
        return nodes


from abc import ABC, abstractmethod

from typing import Dict
import importlib.util
import sys
from pathlib import Path
from orca.helper import FilepathReconciler
from orca.system.system import ISystem
from orca.workflow_models.labware_thread import LabwareThread
from orca.workflow_models.labware_thread import IThreadObserver
from orca.yml_config_builder.configs import ScriptBaseConfigModel


class IThreadScript(IThreadObserver, ABC):
    @abstractmethod
    def thread_notify(self, event: str, thread: LabwareThread) -> None:
        raise NotImplementedError
    

class ThreadScript(IThreadScript, ABC):

    def __init__(self, system: ISystem) -> None:
        super().__init__()
        self._system: ISystem = system

    @property
    def system(self) -> ISystem:
        return self._system

    @abstractmethod
    def thread_notify(self, event: str, thread: LabwareThread) -> None:
        raise NotImplementedError

class IScriptRegistry(ABC):
    @abstractmethod
    def set_system(self, system: ISystem) -> None:
        raise NotImplementedError
        
    @abstractmethod
    def get_script(self, script_name: str) -> ThreadScript:
        raise NotImplementedError

    @abstractmethod
    def add_script(self, name: str, script: ThreadScript) -> None:
        raise NotImplementedError 

    @abstractmethod
    def create_script(self, filepath: str, class_name: str) -> ThreadScript:
        raise NotImplementedError
    
    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError

class IScriptFactory(ABC):
    
    def create_script(self, filepath: str, class_name: str) -> ThreadScript:
        raise NotImplementedError

    def set_system(self, system: ISystem) -> None:
        raise NotImplementedError


class NullScriptFactory(IScriptFactory):

    def create_script(self, filepath: str, class_name: str) -> ThreadScript:
        raise NotImplementedError

    def set_system(self, system: ISystem) -> None:
        self._system = system


class ScriptFactory(IScriptFactory):

    def __init__(self, scripting_config: ScriptBaseConfigModel | None, filepath_reconciler: FilepathReconciler) -> None:
        self._scripting_config = scripting_config
        self._filepath_reconciler = filepath_reconciler
        if self._scripting_config is not None:
            self._filepath_reconciler.set_base_dir(self._scripting_config.base_dir)

    def set_system(self, system: ISystem) -> None:
        self._system = system

    def create_script(self, filepath: str, class_name: str) -> ThreadScript:
        filepath = self._filepath_reconciler.reconcile_filepath(filepath)     

        # get just the filename without the extension
        module_name= Path(filepath).stem
        
        spec = importlib.util.spec_from_file_location(module_name, filepath)
        if spec is None:
                raise ValueError(f"spec not found for {filepath}")

        module_type = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module_type
        if spec.loader is None:
            raise ValueError(f"Loader not found for {filepath}")
        spec.loader.exec_module(module_type)
        try:
            ScriptClass = getattr(module_type, class_name)
        except AttributeError:
            raise ValueError(f"Class {class_name} not found in {filepath}")
        if not issubclass(ScriptClass, ThreadScript):
            raise ValueError(f"Class {class_name} is not a subclass of ThreadScript")
        script_instance: ThreadScript = ScriptClass(self._system)
        return script_instance
    


    
class ScriptRegistry(IScriptRegistry):
    def __init__(self, script_factory: IScriptFactory) -> None:
        self._factory = script_factory
        self._scripts: Dict[str, ThreadScript] = {}

    def set_system(self, system: ISystem) -> None:
        self._factory.set_system(system)

    def get_script(self, script_name: str) -> ThreadScript:
        return self._scripts[script_name]
    
    def add_script(self, name: str, script: ThreadScript) -> None:
        if name in self._scripts.keys():
            raise KeyError(f"Script {name} is already defined in the system.  Each script must have a unique name")
        self._scripts[name] = script

    def create_script(self, filepath: str, class_name: str) -> ThreadScript:
        return self._factory.create_script(filepath, class_name)
    
    def clear(self) -> None:
        self._scripts.clear()
       




import argparse
from orca.cli.orca_cli import OrcaCmdShell
from orca.cli.orca_rest_api import uvicorn_server

def main():
    parser = argparse.ArgumentParser(prog="orca", description="Orca CLI Tool")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Interactive shell subcommand
    subparsers.add_parser("interactive", help="Enter interactive shell")

    # Server subcommand
    subparsers.add_parser("server", help="Start Orca server")

    # No subcommand means direct CLI command execution
    parser.add_argument("args", nargs=argparse.REMAINDER, help="Command to execute")

    # Parse the arguments
    args = parser.parse_args()

    if args.command == "interactive":
        OrcaCmdShell().cmdloop()
    elif args.command == "server":
        uvicorn_server.run()
    elif args.args:
        # Execute the command using OrcaCmdShell's onecmd
        OrcaCmdShell().onecmd(" ".join(args.args))
    else:
        parser.print_help()

if __name__ == "__main__":
    uvicorn_server.run()


import asyncio
import logging
from orca.resource_models.base_resource import Equipment
from orca_driver_interface.transporter_interfaces import ITransporterDriver
from orca.resource_models.location import Location
from typing import List, Optional
from orca.resource_models.labware import Labware

orca_logger = logging.getLogger("orca")

class TransporterEquipment(Equipment):
    def __init__(self, name: str, driver: ITransporterDriver) -> None:
        super().__init__(name, driver)
        self._driver: ITransporterDriver = driver
        self._labware: Optional[Labware] = None
        self._lock = asyncio.Lock()

    @property
    def labware(self) -> Optional[Labware]:
        return self._labware
    
    async def pick(self, location: Location) -> None:
        async with self._lock:
            if self._labware is not None:
                raise ValueError(f"{self} already contains labware: {self._labware}")
            if location.labware is None:
                raise ValueError(f"{location} does not contain labware")
            orca_logger.info(f"{self._name} pick {location.labware} from {location}: picking...")
            await self._driver.pick(location.teachpoint_name, location.labware.labware_type)
            orca_logger.info(f"{self._name} pick {location.labware} from {location}: picked")
            self._labware = location.labware

    async def place(self, location: Location) -> None:
        async with self._lock:
            if self._labware is None:
                raise ValueError(f"{self} does not contain labware")
            if location.labware is not None:
                raise ValueError(f"{location} already contains labware")
            orca_logger.info(f"{self._name} place {self._labware} to {location}: placing...")
            await self._driver.place(location.teachpoint_name, self._labware.labware_type)
            orca_logger.info(f"{self._name} place {self._labware} to {location}: placed")
            
            self._labware = None

    def get_taught_positions(self) -> List[str]:
        return self._driver.get_taught_positions()
    
    def __str__(self) -> str:
        return f"Transporter: {self._name}"
    


from orca.resource_models.labware import Labware, LabwareTemplate

from abc import ABC, abstractmethod
from typing import List


class ILabwareRegistry(ABC):

    @property
    @abstractmethod
    def labwares(self) -> List[Labware]:
        pass

    @abstractmethod
    def get_labware(self, name: str) -> Labware:
        raise NotImplementedError

    @abstractmethod
    def add_labware(self, labware: Labware) -> None:
        raise NotImplementedError


class ILabwareTemplateRegistry(ABC):
    @abstractmethod
    def get_labware_template(self, name: str) -> LabwareTemplate:
        raise NotImplementedError

    @abstractmethod
    def add_labware_template(self, labware: LabwareTemplate) -> None:
        raise NotImplementedError